from .account import RoleSyncService, UserRoleSyncService, UserSyncService
from .competition import (
    FixtureService,
    LeaguePropertiesService,
    LeagueService,
    OrganizationService,
    PhaseService,
    RulesetService,
    SeasonService,
    SiteService,
)
from .membership import (
    AthleteService,
    IndividualMembershipService,
    ManagerMembershipService,
    ManagerService,
    OfficialService,
    TeamMembershipService,
    TeamService,
)
from .participation import AthleteStatsService, ManagingService, OfficiatingService, TeamStatsService

__all__ = [
    "RoleSyncService",
    "UserRoleSyncService",
    "UserSyncService",
    "FixtureService",
    "LeaguePropertiesService",
    "LeagueService",
    "OrganizationService",
    "PhaseService",
    "RulesetService",
    "SeasonService",
    "SiteService",
    "AthleteService",
    "AthleteStatsService",
    "IndividualMembershipService",
    "ManagerMembershipService",
    "ManagerService",
    "OfficialService",
    "TeamMembershipService",
    "TeamService",
    "ManagingService",
    "OfficiatingService",
    "TeamStatsService",
]



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Team
from leaguemanager.repository import TeamSyncRepository
from leaguemanager.repository._async import TeamAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["TeamService", "TeamAsyncService"]


class TeamService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = TeamSyncRepository


class TeamAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = TeamAsyncRepository



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import SiteBase, mapper, metadata

if TYPE_CHECKING:
    from .league import League


@define(slots=False)
class Site(SiteBase):
    """Defines a specific location.

    A Site can be assigned to Organization, League, Season, or Fixture base
    objects. This is because different sites can exist based on the matchups.

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        organization_id (UUID): ForeignKey to Organization
        league_id (UUID): ForeignKey to League.
        season_id (UUID): ForeignKey to Season
        name (str): Descriptive name. (i.e, Barnes Center, The Pinback Complex)
        description (str): Optional description of the site.
        specifier (str): If specific field/rink/area to be designated (i.e., Field C, Gym Court 2, etc...)
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    organization_id: UUID | None = field(default=None)
    league_id: UUID | None = field(default=None)
    season_id: UUID | None = field(default=None)
    name: str | None = field(default=None, validator=validators.max_len(40))
    description: str | None = field(default=None, validator=validators.max_len(100))
    specifier: str | None = field(default=None, validator=validators.max_len(12))


# SQLAlchemy Imperative Mappings

site = Table(
    "site",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("organization_id", SA_UUID, ForeignKey("organization.id"), nullable=True),
    Column("league_id", SA_UUID, ForeignKey("league.id"), nullable=True),
    Column("season_id", SA_UUID, ForeignKey("season.id"), nullable=True),
    Column("name", String(40), nullable=True),
    Column("description", String(100), nullable=True),
    Column("specifier", String(12), nullable=True),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    Site,
    site,
    properties={
        "organization": relationship("Organization", back_populates="site"),
        "league": relationship("League", back_populates="site"),
        "season": relationship("Season", back_populates="site"),
        "ruleset": relationship("Ruleset", back_populates="site"),
        "fixture": relationship("Fixture", back_populates="site", uselist=False),
        "address": relationship("Address", back_populates="site"),
    },
)



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Column, DateTime, ForeignKey, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import EventBase, mapper, metadata


@define(slots=False)
class Fixture(EventBase):
    """Defines a contest/competition which results in statistical results.

    A Fixture represents a specific contest or match between opponents or contestants.
    This object is generally a child of a CompetitionPhase object. For example, during any given
    phase of a competition (such as a Matchday), there can be many Fixtures (i.e., Team A vs Team B).

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        season_id (UUID): ForeignKey to Season.
        phase_id (UUID): ForeignKey to Phase.
        site_id (UUID): ForeignKey to Site.
        label (str): Description of the fixture (i.e., Team A vs Team B, 8-Team Elimination, etc...)
        outcome_type (str): Type of outcome (i.e., regular, decision, shootout, forfeit, etc...)
        status (str): Status of fixture (i.e., pre-event, in-progress, post-event).
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    season_id: UUID | None = field(default=None)
    phase_id: UUID | None = field(default=None)
    site_id: UUID | None = field(default=None)
    label: str | None = field(default=None, validator=validators.max_len(40))
    outcome_type: str | None = field(default="regular", validator=validators.max_len(12))
    status: str | None = field(default="pre-event", validator=validators.max_len(12))


# SQLAlchemy Imperative Mappings

fixture = Table(
    "fixture",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("season_id", SA_UUID, ForeignKey("season.id"), nullable=False),
    Column("phase_id", SA_UUID, ForeignKey("phase.id"), nullable=True),
    Column("site_id", SA_UUID, ForeignKey("site.id"), nullable=True),
    Column("label", String(40), nullable=False),
    Column("outcome_type", String(12), nullable=True),
    Column("status", String(12), nullable=True),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    Fixture,
    fixture,
    properties={
        "season": relationship("Season", back_populates="fixtures"),
        "phase": relationship("Phase", back_populates="fixtures"),
        "team_stats": relationship("TeamStats", back_populates="fixture"),
        "athlete_stats": relationship("AthleteStats", back_populates="fixture"),
        "managing": relationship("Managing", back_populates="fixture"),
        "officiating": relationship("Officiating", back_populates="fixture"),
        "site": relationship("Site", back_populates="fixture", uselist=False),
    },
)



import typer
from advanced_alchemy.cli import add_migration_commands

from leaguemanager.dependency.cli_callbacks import provide_sync_db_config

app = typer.Typer(name="db", help="Database commands.", no_args_is_help=True)


@app.callback(context_settings={"obj": {"configs": [provide_sync_db_config()]}})
def _app(ctx: typer.Context):
    """Passes SQLAlchemy configuration to the Click database group within Advanced Alchemy.

    The SQLAlchemySyncConfig and SQLAlchemyAsyncConfig classes that are defined elsewhere are
    imported here and passed through the Click context. The configuration is accessed by the
    Advanced Alchemy migration commands.

    Args:
        ctx (typer.Context): The Click context.
    """
    pass


# Generate a Click group from Typer "db" app and add Advanced Alchemy migration commands.
click_group = typer.main.get_command(app)

# Add the db_app to the main Typer app to access the db commands.
db_app = add_migration_commands(click_group)



from .athlete_stats import AthleteStats
from .managing import Managing
from .officiating import Officiating
from .team_stats import TeamStats

__all__ = [
    "AthleteStats",
    "Managing",
    "Officiating",
    "TeamStats",
]



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Column, DateTime, ForeignKey, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import GoverningBodyBase, mapper, metadata

if TYPE_CHECKING:
    from .league import League


@define(slots=False)
class Organization(GoverningBodyBase):
    """Defines the parent organization that administers competition(s).

    This is the top-level object that defines the organization responsible for the rules and
    management of leagues and/or competitions. Generally will have CompetitionBase children
    to describe what kind of competition is being administered.

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        org_type (str): Type of organization (i.e., amateur, youth, etc...). Inherited from GoverningBodyBase.
        name (str): Descriptive name. (i.e., Coed Hockey League, Flag Football at the Park, etc...)
        description (str): Optional description of the league.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    name: str | None = field(default=None, validator=validators.max_len(100))
    description: str | None = field(default=None, validator=validators.optional(validators.max_len(255)))


organization = Table(
    "organization",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("org_type", String(20), nullable=True),
    Column("name", String(80), nullable=False, unique=True),
    Column("description", String(255), nullable=True),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships
mapper.map_imperatively(
    Organization,
    organization,
    properties={
        "leagues": relationship("League", back_populates="organization"),
        "site": relationship("Site", back_populates="organization"),
    },
)



# Configuration file for the Sphinx documentation builder.
# ruff: noqa: FIX002 PLR0911 ARG001 ERA001

__version__ = "0.8.1"


project = "League Manager"
copyright = "2025, Mario Munoz"
author = "Mario Munoz"
release = __version__
suppress_warnings = [
    "autosectionlabel.*",
    "ref.python",  # remove when https://github.com/sphinx-doc/sphinx/issues/4961 is fixed
]

# -- General configuration ---------------------------------------------------

extensions = [
    "autodoc2",
    "myst_parser",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_design",
    "sphinx_copybutton",
    "sphinx_toolbox.collapse",
    "sphinx_togglebutton",
]

myst_enable_extensions = [
    "attrs_inline",
    "colon_fence",
    "fieldlist",
    "html_admonition",
    "html_image",
    "replacements",
    "smartquotes",
    "strikethrough",
    "tasklist",
]

autodoc2_render_plugin = "myst"
autodoc2_packages = [
    {
        "path": "../src/leaguemanager",
        "exclude_files": [r"__.*\.py$"],
    },
]

napoleon_google_docstring = True
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = False
napoleon_attr_annotations = True

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

html_theme = "piccolo_theme"
html_title = "League Manager"
html_short_title = "leaguemanager"
templates_path = ["_templates"]
html_static_path = ["_static"]

html_theme_options = {"source_url": "https://codeberg.org/pythonbynight/leaguemanager", "source_icon": "generic"}
html_css_files = ["custom.css"]



from typing import Any

import attrs
from advanced_alchemy.service import SQLAlchemyAsyncRepositoryService as _SQLAlchemyAsyncRepositoryService
from advanced_alchemy.service import SQLAlchemySyncRepositoryService as _SQLAlchemySyncRepositoryService

from ._typing import ModelT


class SQLAlchemySyncRepositoryService(_SQLAlchemySyncRepositoryService):
    """Introduces a `to_model` method to transform ModelT data to a dictionary. This is
    only necessary because SQLAlchemySyncRepositoryService does not work out of the box
    for attrs defined classes.
    """

    def to_model(
        self,
        data: ModelT | dict[str, Any],
        operation: str | None = None,
    ) -> ModelT:
        if attrs.has(data):
            data = attrs.asdict(data)
        return super().to_model(data, operation)


# Not sure if this is needed
# TODO: Seems like litestar checks if data received is a certain kind before transforming "to model"
class SQLAlchemyAsyncRepositoryService(_SQLAlchemyAsyncRepositoryService):
    """Introduces a `to_model` method to transform ModelT data to a dictionary. This is
    only necessary because SQLAlchemyAsyncRepositoryService does not work out of the box
    for attrs defined classes.
    """

    async def to_model(
        self,
        data: ModelT | dict[str, Any],
        operation: str | None = None,
    ) -> ModelT:
        if attrs.has(data):
            data = attrs.asdict(data)
        return await super().to_model(data, operation)



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import ParticipationBase, mapper, metadata

if TYPE_CHECKING:
    from leaguemanager.models.baseline import Team


@define(slots=False)
class TeamStats(ParticipationBase):
    """The participation of an individual or team that can win or lose an Fixture.

    Each team entity that is participating in an Fixture has a TeamStats that represents
    where scores and stats are expressed about the team's performance in that
    given Fixture.

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        season_id (UUID): ForeignKey to Season.
        fixture_id (UUID): ForeignKey to Fixture.
        label (str): Description of TeamStats (i.e., "Team A, Week 2, Sunday Softball Div 2", etc...)
        outcome (str): Whether result of Team was a "win", "loss", "tie", or "undecided".
        score_for (int): The total points scored by the Team.
        score_against (int): The total poinst scored against the Team.
        alignment (str): Whether the Team is "home" or "away" (one of those values).
        fixture_standing_points (int): The number of points from the results of this fixture.
        standing_points (int): The number of total points during the competition.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    season_id: UUID | None = field(default=None)
    fixture_id: UUID | None = field(default=None)
    label: str | None = field(default=None, validator=validators.max_len(90))
    outcome: str | None = field(
        default=None, validator=validators.optional(validators.in_(("win", "loss", "tie", "undecided")))
    )
    score_for: int | None = field(default=0)
    score_against: int | None = field(default=0)
    alignment: str | None = field(default=None, validator=validators.optional(validators.in_(("home", "away"))))
    fixture_standing_points: int | None = field(default=None)
    standing_points: int | None = field(default=None)


# SQLAlchemy Imperative Mappings

team_stats = Table(
    "team_stats",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("season_id", SA_UUID, ForeignKey("season.id"), nullable=False),
    Column("fixture_id", SA_UUID, ForeignKey("fixture.id"), nullable=True),
    Column("label", String(90), nullable=False),
    Column("outcome", String(12), nullable=True),
    Column("score_for", Integer, nullable=True),
    Column("score_against", Integer, nullable=True),
    Column("alignment", String(6), nullable=True),
    Column("fixture_standing_points", Integer, nullable=True),
    Column("standing_points", Integer, nullable=True),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    TeamStats,
    team_stats,
    properties={
        "season": relationship("Season", back_populates="team_stats"),
        "fixture": relationship("Fixture", back_populates="team_stats"),
        "team": relationship("Team", back_populates="team_stats", uselist=False),
    },
)



from contextlib import asynccontextmanager, contextmanager
from typing import Any, AsyncGenerator, Generator

from advanced_alchemy.config import AlembicAsyncConfig, AlembicSyncConfig, SQLAlchemyAsyncConfig, SQLAlchemySyncConfig
from sqlalchemy import create_engine
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.orm import Session, sessionmaker

from leaguemanager.core import get_settings

from .db_config import db_args, uri

__all__ = [
    "sync_alembic_config",
    "sync_engine",
    "sync_session_factory",
    "sync_config",
    "async_alembic_config",
    "async_engine",
    "async_session_factory",
    "async_config",
    "get_session",
    "get_async_session",
]


settings = get_settings()

# Sync DB Setup

sync_alembic_config = AlembicSyncConfig(
    version_table_name="alembic_version",
    script_config=str(settings.MIGRATION_CONFIG / "alembic.ini"),
    script_location=str(settings.MIGRATION_PATH),
    template_path=settings.ALEMBIC_TEMPLATE_PATH,
)
sync_engine = create_engine(url=uri(), **db_args)
sync_session_factory = sessionmaker(bind=sync_engine, expire_on_commit=False)
sync_config = SQLAlchemySyncConfig(
    engine_instance=sync_engine, session_maker=sync_session_factory, alembic_config=sync_alembic_config
)


@contextmanager
def get_session() -> Generator[Session, Any, None]:
    with sync_session_factory() as session:
        yield session


# Async DB Setup

async_alembic_config = AlembicAsyncConfig(
    version_table_name="alembic_version",
    script_config=str(settings.MIGRATION_CONFIG / "alembic.ini"),
    script_location=str(settings.MIGRATION_PATH),
    template_path=settings.ALEMBIC_TEMPLATE_PATH,
)
async_engine = create_async_engine(url=uri(is_async=True), **db_args)
async_session_factory = async_sessionmaker(async_engine, expire_on_commit=False)
async_config = SQLAlchemyAsyncConfig(
    create_engine_callable=create_async_engine,
    engine_instance=async_engine,
    session_maker=async_session_factory,
    alembic_config=async_alembic_config,
)


@asynccontextmanager
async def get_async_session() -> AsyncGenerator[AsyncSession, None]:
    async with async_session_factory() as session:
        yield session


def validate_config(config: SQLAlchemySyncConfig | SQLAlchemyAsyncConfig, is_async: bool = False) -> None:
    if not config.engine_instance:
        raise ValueError("Engine instance is not set")
    if not config.session_maker:
        raise ValueError("Session maker is not set")
    if not config.alembic_config:
        if is_async:
            config.alembic_config = async_alembic_config
        else:
            config.alembic_config = sync_alembic_config
    return config



from datetime import UTC, datetime

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import IndividualBase, mapper, metadata

if TYPE_CHECKING:
    from .user_role import UserRole


@define(slots=False)
class User(IndividualBase):
    """A user of the system."""

    username: str | None = field(default=None, validator=validators.optional(validators.max_len(40)))
    email: str | None = field(default=None, validator=validators.max_len(40))
    password: str | None = field(default=None, validator=validators.max_len(32))
    active: bool = field(default=True)
    superuser: bool = field(default=False)
    joined: datetime | None = field(factory=lambda: datetime.now(UTC))


# SQLAlchemy Imperative Mapping

user = Table(
    "user",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("username", String(40), unique=True, nullable=True),
    Column("email", String(40), unique=True, nullable=False),
    Column("password", String(32), nullable=False),
    Column("active", Boolean, default=True),
    Column("superuser", Boolean, default=False),
    Column("joined", DateTime, default=lambda: datetime.now(UTC)),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships

mapper.map_imperatively(
    User,
    user,
    properties={
        "user_roles": relationship("UserRole", back_populates="user"),
        # "admin": relationship("Admin", back_populates="user"),
    },
)



import asyncio
from logging.config import fileConfig
from typing import TYPE_CHECKING

from alembic import context
from sqlalchemy import engine_from_config, pool

from leaguemanager.db.db_config import uri
from leaguemanager.models.base import metadata

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata

target_metadata = metadata
# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """

    context.configure(
        url=uri(is_async=True),
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def do_run_migrations(connection: "Connection") -> None:
    """Run migrations."""
    context.configure(
        url=uri(is_async=True),
        connection=connection,
        target_metadata=target_metadata,
        compare_type=True,
        render_as_batch=True,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine and associate a
    connection with the context.
    """

    configuration = config.get_section(config.config_ini_section) or {}
    configuration["sqlalchemy.url"] = uri()
    connectable = engine_from_config(
        configuration,
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
        future=True,
    )

    if connectable is None:  # pyright: ignore[reportUnnecessaryComparison]
        msg = "Could not get engine from config.  Please ensure your `alembic.ini` according to the official Alembic documentation."
        raise RuntimeError(
            msg,
        )

    with connectable.connect() as connection:
        do_run_migrations(connection=connection)

    connectable.dispose()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from leaguemanager import models as m
from leaguemanager import services as s
from leaguemanager.db import register_sqlite
from leaguemanager.models.base import metadata

pytest_plugins = ["tests.data_fixtures"]

register_sqlite()


@pytest.fixture(scope="session")
def db_dir(tmp_path_factory):
    db_dir = tmp_path_factory.mktemp("db_test")
    return db_dir


# Sync DB fixtures


@pytest.fixture(scope="session")
def engine(db_dir):
    uri = f"sqlite:///{db_dir / 'lmgr_db.db'}"
    engine = create_engine(uri, echo=False)
    metadata.drop_all(engine)
    metadata.create_all(engine)
    yield engine
    metadata.drop_all(engine)
    engine.dispose()


@pytest.fixture(scope="session")
def session_factory(engine):
    return sessionmaker(bind=engine, expire_on_commit=True, autocommit=False, autoflush=False)


@pytest.fixture(scope="session", autouse=True)
def session(session_factory, all_data):
    session_ = session_factory()
    yield session_
    session_.rollback()
    session_.close()


@pytest.fixture(scope="session")
def org_service(session):
    return s.OrganizationService(session=session)


@pytest.fixture(scope="session")
def league_service(session):
    return s.LeagueService(session=session)


@pytest.fixture(scope="session")
def league_properties_service(session):
    return s.LeaguePropertiesService(session=session)


@pytest.fixture(scope="session")
def season_service(session):
    return s.SeasonService(session=session)


@pytest.fixture(scope="session")
def ruleset_service(session):
    return s.RulesetService(session=session)


@pytest.fixture(scope="session")
def fixture_service(session):
    return s.FixtureService(session=session)


@pytest.fixture(scope="session")
def team_service(session):
    return s.TeamService(session=session)


@pytest.fixture(scope="session")
def team_member_service(session):
    return s.TeamMembershipService(session=session)


@pytest.fixture(scope="session")
def athlete_service(session):
    return s.AthleteService(session=session)


@pytest.fixture(scope="session")
def individual_membership_service(session):
    return s.IndividualMembershipService(session=session)



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import MembershipBase, mapper, metadata


@define(slots=False)
class IndividualMembership(MembershipBase):
    """An individual's membership in a Team.

    Instead of linking an Athlete object directly to a team, the Athlete is linked
    to an IndividualMembership object, which defines their specific affiliation to a
    given team. That way, an Athlete can be linked to different IndividualMembership
    objects, each connected to different Teams (in the event that an Athlete has multiple
    team affiliations). The IndividualMembership object can define properties specific
    to a Team, such as uniform number, start and end dates, etc...

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        team_id (UUID): ForeignKey to Team.
        athlete_id (UUID): ForeignKey to Athlete.
        label (str): Label describing the Athlete association (i.e., Henry Cooper, Team A, etc...)
        uniform_number (int): Number used for uniform.
        start_date (datetime): Date that individual joined Team.
        end_date (datetime): Date that individual left Team.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    team_id: UUID | None = field(default=None)
    athlete_id: UUID | None = field(default=None)
    label: str | None = field(default=None, validator=validators.optional(validators.max_len(24)))
    uniform_number: int | None = field(default=None)
    start_date: datetime | None = field(default=None)
    end_date: datetime | None = field(default=None)


# SQLAlchemy Imperative Mapping

individual_membership = Table(
    "individual_membership",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("team_id", SA_UUID, ForeignKey("team.id"), nullable=True),
    Column("athlete_id", SA_UUID, ForeignKey("athlete.id"), nullable=True),
    Column("label", String(24), nullable=True),
    Column("uniform_number", Integer, nullable=True),
    Column("start_date", DateTime(), nullable=True, default=None),
    Column("end_date", DateTime(), nullable=True, default=None),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    IndividualMembership,
    individual_membership,
    properties={
        "athlete": relationship("Athlete", back_populates="individual_membership", uselist=False),
        "team": relationship("Team", back_populates="individual_memberships"),
    },
)



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import AgentBase, mapper, metadata


@define(slots=False)
class Team(AgentBase):
    """A group of individuals competing as one entitiy.

    Although it may be unusual, it is sometimes possible for a Team to have
    "membership" in different Leagues. For this reason, a Team is not directly
    linked to a League, but rather, to a TeamMebership object. The TeamMembership
    object will then define the League (or Leagues) a Team belongs to, and can
    also track start/end dates for said membership.

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        team_membership_id (UUID): ForeignKey to TeamMembership.
        team_stats_id (UUID): ForeignKey to TeamStats.
        name (str): Team name.
        active (bool): Defaults to True. Can be used to deactivate this Team.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    team_membership_id: UUID | None = field(default=None)
    team_stats_id: UUID | None = field(default=None)
    name: str | None = field(default=None, validator=validators.max_len(40))
    active: bool = field(default=True)


# SQLAlchemy Imperative Mappings

team = Table(
    "team",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("team_membership_id", SA_UUID, ForeignKey("team_membership.id"), nullable=True),
    Column("team_stats_id", SA_UUID, ForeignKey("team_stats.id"), nullable=True),
    Column("name", String(40), nullable=False),
    Column("active", Boolean, default=True),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships

mapper.map_imperatively(
    Team,
    team,
    properties={
        "team_membership": relationship("TeamMembership", back_populates="team"),
        "individual_memberships": relationship("IndividualMembership", back_populates="team"),
        "manager_memberships": relationship("ManagerMembership", back_populates="team"),
        "team_stats": relationship("TeamStats", back_populates="team"),
    },
)



from .competition import Fixture, League, LeagueProperties, Organization, Phase, Ruleset, Season, Site
from .membership import Athlete, IndividualMembership, Manager, ManagerMembership, Official, Team, TeamMembership
from .participation import AthleteStats, Managing, Officiating, TeamStats

__all__ = [
    "Fixture",
    "League",
    "LeagueProperties",
    "Organization",
    "Phase",
    "Ruleset",
    "Season",
    "Site",
    "Athlete",
    "IndividualMembership",
    "Manager",
    "ManagerMembership",
    "Official",
    "Team",
    "TeamMembership",
    "AthleteStats",
    "Managing",
    "Officiating",
    "TeamStats",
]



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import User, UserRole
from leaguemanager.repository import UserRoleSyncRepository, UserSyncRepository
from leaguemanager.repository._async import UserAsyncRepository, UserRoleAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["UserSyncService", "UserAsyncService", "UserRoleSyncService", "UserRoleAsyncService"]


class UserSyncService(SQLAlchemySyncRepositoryService):
    """Handles user database operations."""

    repository_type = UserSyncRepository


class UserRoleSyncService(SQLAlchemySyncRepositoryService):
    """Handles database operations in the user/role association."""

    repository_type = UserRoleSyncRepository


class UserAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles user database operations."""

    repository_type = UserAsyncRepository


class UserRoleAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations in the user/role association."""

    repository_type = UserRoleAsyncRepository



from .athlete_stats import AthleteStatsAsyncService, AthleteStatsService
from .managing import ManagingAsyncService, ManagingService
from .officiating import OfficiatingAsyncService, OfficiatingService
from .team_stats import TeamStatsAsyncService, TeamStatsService

__all__ = [
    "AthleteStatsService",
    "AthleteStatsAsyncService",
    "ManagingService",
    "ManagingAsyncService",
    "OfficiatingService",
    "OfficiatingAsyncService",
    "TeamStatsService",
    "TeamStatsAsyncService",
]



from .account import RoleAsyncService, UserAsyncService, UserRoleAsyncService
from .competition import (
    FixtureAsyncService,
    LeagueAsyncService,
    LeaguePropertiesAsyncService,
    OrganizationAsyncService,
    PhaseAsyncService,
    RulesetAsyncService,
    SeasonAsyncService,
    SiteAsyncService,
)
from .membership import (
    AthleteAsyncService,
    IndividualMembershipAsyncService,
    ManagerAsyncService,
    ManagerMembershipAsyncService,
    OfficialAsyncService,
    TeamAsyncService,
    TeamMembershipAsyncService,
)
from .participation import (
    AthleteStatsAsyncService,
    ManagingAsyncService,
    OfficiatingAsyncService,
    TeamStatsAsyncService,
)

__all__ = [
    "RoleAsyncService",
    "UserAsyncService",
    "UserRoleAsyncService",
    "FixtureAsyncService",
    "LeagueAsyncService",
    "LeaguePropertiesAsyncService",
    "OrganizationAsyncService",
    "PhaseAsyncService",
    "RulesetAsyncService",
    "SeasonAsyncService",
    "SiteAsyncService",
    "AthleteAsyncService",
    "AthleteStatsAsyncService",
    "IndividualMembershipAsyncService",
    "ManagerAsyncService",
    "ManagerMembershipAsyncService",
    "OfficialAsyncService",
    "TeamAsyncService",
    "TeamMembershipAsyncService",
    "ManagingAsyncService",
    "OfficiatingAsyncService",
    "TeamStatsAsyncService",
]



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Phase
from leaguemanager.repository import PhaseSyncRepository
from leaguemanager.repository._async import PhaseAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["PhaseService", "PhaseAsyncService"]


class PhaseService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = PhaseSyncRepository


class PhaseAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = PhaseAsyncRepository



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import MembershipBase, mapper, metadata


@define(slots=False)
class TeamMembership(MembershipBase):
    """A teams membership in a given Competition (i.e., League, Season).

    Instead of linking a Team object directly to a Competition, the Team is linked
    to a TeamMembership object, which defines their specific affiliation to a
    given competition. That way, a Team can be linked to different TeamMembership
    objects, each connected to different Competitions (in the event that an Team belongs
    to differet Seasons/Leagues). The TeamMembership object can define properties specific
    to a Team, such as uniform color, start and end dates, etc...

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        season_id (UUID): ForeignKey to Season.
        label (str): Label describing the Team association (i.e., Cylon FC--Fall 2024, etc...)
        home_color (str): Primary color of home uniform.
        away_color (str): Primary color of away uniform.
        start_date (datetime): Date that individual joined Team.
        end_date (datetime): Date that individual left Team.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    season_id: UUID | None = field(default=None)
    label: str | None = field(default=None, validator=validators.optional(validators.max_len(80)))
    home_color: str | None = field(default="", validator=validators.max_len(12))
    away_color: str | None = field(default="", validator=validators.max_len(12))
    start_date: datetime | None = field(default=None)
    end_date: datetime | None = field(default=None)


# SQLAlchemy Imperative Mappings

team_membership = Table(
    "team_membership",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("season_id", SA_UUID, ForeignKey("season.id"), nullable=True),
    Column("label", String(80), nullable=False),
    Column("home_color", String(12), nullable=True),
    Column("away_color", String(12), nullable=True),
    Column("start_date", DateTime(), nullable=True, default=None),
    Column("end_date", DateTime(), nullable=True, default=None),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    TeamMembership,
    team_membership,
    properties={
        "team": relationship("Team", back_populates="team_membership", uselist=False),
        "season": relationship("Season", back_populates="team_memberships"),
    },
)



from datetime import datetime, timedelta
from uuid import uuid4

import pytest
from advanced_alchemy.filters import CollectionFilter, OnBeforeAfter
from sqlalchemy.exc import NoResultFound

from leaguemanager.core.toolbox import clear_table
from leaguemanager.models import Fixture, Team
from leaguemanager.services import (
    FixtureService,
    LeagueService,
    SeasonService,
    TeamService,
)


@pytest.fixture(scope="module")
def season(season_service) -> SeasonService:
    season = season_service.create(
        data={"name": "SZNZ", "description": "Winter", "projected_start_date": "2022-01-01"}, auto_commit=True
    )
    return season


@pytest.fixture(scope="module")
def league(league_service, season) -> LeagueService:
    league = league_service.create(data={"name": "Men's 7v7", "category": "MEN", "division": "A", "match_day": "SUN"})
    league.season = season
    league = league_service.update(league, auto_commit=True)
    return league


@pytest.fixture(scope="module")
def teams(team_service, league_service, league) -> TeamService:
    teams = team_service.create_many(
        [
            {"name": "Team 1", "league_id": league.id},
            {"name": "Team 2", "league_id": league.id},
            {"name": "Team 3", "league_id": league.id},
            {"name": "Team 4", "league_id": league.id},
        ]
    )
    league.teams = teams
    league_service.update(league, auto_commit=True)
    return teams


# @pytest.fixture(scope="module")
# def schedule(schedule_service, league, season, teams) -> ScheduleService:
#     _schedule = {
#         "name": "I Want A Dog",
#         "total_games": 10,
#         "game_length": 45,
#         "time_between_games": 15,
#         "start_date": "2025-1-05 16:00:00",
#         "season_id": season.id,
#         "league_id": league.id,
#         "teams": teams,
#     }

#     schedule = schedule_service.create(_schedule)
#     return schedule


# @pytest.fixture(scope="module")
# def scheduler(session, schedule) -> Scheduler:
#     scheduler = Scheduler(session=session, schedule=schedule)
#     return scheduler


# @pytest.fixture(scope="module")
# def league_fixtures(fixture_service, scheduler) -> FixtureService:
#     fixtures = scheduler.generate_fixtures(shuffle_teams=False)
#     fixture_service.create_many(fixtures, auto_commit=True)
#     return fixtures


# @pytest.fixture(scope="module")
# def standings_table(session, schedule):
#     standings_table = StandingsTable(
#         session=session,
#         schedule=schedule,
#     )

#     return standings_table


# def test_standings_table_properties(schedule, standings_table):
#     assert standings_table.schedule == schedule
#     assert standings_table.schedule.season == schedule.season
#     assert standings_table.schedule.season.name == "SZNZ"


# def test_get_or_create_standings(standings_table, standings_service):
#     standings = standings_table.get_or_create_current_standings()
#     assert len(standings) == 4

#     for standing in standings:
#         standing.games_played = 1
#         standings_service.update(standing, auto_commit=True)

#     new_standings = standings_table.get_or_create_current_standings()
#     assert len(new_standings) == 4
#     assert new_standings[0].games_played == 1
#     assert new_standings[1].games_played == 1
#     assert new_standings[2].games_played == 1
#     assert new_standings[3].games_played == 1


# def test_past_fixtures_all_unplayed(standings_table, league_fixtures, fixture_service):
#     fixtures = standings_table.past_fixtures
#     assert len(fixtures) == 0

#     _fixture = league_fixtures[0]
#     _fixture.status = "P"
#     fixture_service.update(_fixture, auto_commit=True)

#     fixtures = standings_table.past_fixtures
#     assert len(fixtures) == 1
#     _fixture = league_fixtures[0]
#     _fixture.status = "U"
#     fixture_service.update(_fixture, auto_commit=True)


# def test_remaining_fixtures_all_unplayed(standings_table, league_fixtures, fixture_service):
#     fixtures = standings_table.remaining_fixtures
#     assert len(fixtures) == 16

#     _fixture = league_fixtures[0]
#     _fixture.status = "P"
#     fixture_service.update(_fixture, auto_commit=True)

#     fixtures = standings_table.remaining_fixtures
#     assert len(fixtures) == 15
#     _fixture = league_fixtures[0]
#     _fixture.status = "U"
#     fixture_service.update(_fixture, auto_commit=True)


# @pytest.mark.parametrize("time_delta_days, expected", [(6, 2), (12, 4), (24, 8)])
# def test_next_scheduled_fixtures(standings_table, league_fixtures, time_delta_days, expected):
#     # Note: league_fixtures needs to be here to generate fixtures for this test
#     # if run individually.
#     all_fixtures = standings_table.all_fixtures
#     assert len(all_fixtures) == 16
#     assert all_fixtures[0].date == datetime(2025, 1, 5, 16, 0)
#     assert all_fixtures[1].date == datetime(2025, 1, 5, 16, 0)

#     filter = OnBeforeAfter(
#         field_name="date",
#         on_or_after=datetime(2025, 1, 4, 16, 0),
#         on_or_before=datetime(2025, 1, 4, 16, 0) + timedelta(days=time_delta_days),
#     )

#     fixtures = standings_table.next_scheduled_fixtures(filter=filter)
#     assert len(fixtures) == expected


# def test_link_teams_to_fixtures(standings_table, league_fixtures):
#     # Note: league_fixtures needs to be here to generate fixtures for this test
#     # if run individually.
#     fixture_teams = standings_table.link_teams_to_fixtures()
#     assert len(fixture_teams) == 32


# def test_link_teams_to_fixtures_when_no_fixtures(session, schedule_service):
#     temp_schedule = Schedule(
#         season_id=uuid4(),
#         league_id=uuid4(),
#         start_date="2021-1-08 18:00:00",
#         end_date="2022-02-01 18:00:00",
#         total_games=0,
#     )
#     temp_standings = StandingsTable(session=session, schedule=temp_schedule)

#     with pytest.raises(NoResultFound):
#         temp_standings.link_teams_to_fixtures()


# def test_team_update_standings(
#     league, session, schedule, standings_service, fixture_team_service, mocker, team_service
# ):
#     mock_home_team = Team(league_id=league.id, name="Mock Home")
#     mock_away_team = Team(league_id=league.id, name="Mock Away")

#     team_service.create(mock_home_team)
#     team_service.create(mock_away_team)

#     fake_fixture = Fixture(
#         home_team_id=mock_home_team.id,
#         away_team_id=mock_away_team.id,
#         schedule_id=uuid4(),
#         status="P",
#         home_goals=3,
#         away_goals=1,
#     )

#     mock_home_stat = FixtureTeam(
#         team_id=mock_home_team.id,
#         fixture_id=fake_fixture.id,
#         is_played=True,
#         points=3,
#         score_for=3,
#         score_against=1,
#         fixture_result="W",
#     )
#     mock_away_stat = FixtureTeam(
#         team_id=mock_away_team.id,
#         fixture_id=fake_fixture.id,
#         is_played=True,
#         points=0,
#         score_for=1,
#         score_against=3,
#         fixture_result="L",
#     )

#     fixture_team_service.create(mock_home_stat)
#     fixture_team_service.create(mock_away_stat)

#     mock_standings = Standings(
#         schedule_id=schedule.id,
#         team_id=mock_home_team.id,
#     )

#     standings_service.create(mock_standings)

#     tabulate_fixture = TabulateFixture(
#         session=session,
#         fixture=fake_fixture,
#     )

#     expected_standings_after_update = Standings(
#         schedule_id=schedule.id,
#         team_id=mock_home_team.id,
#         points=3,
#         games_played=1,
#         games_won=1,
#         games_drawn=0,
#         games_lost=0,
#         score_for=3,
#         score_against=1,
#         score_diff=2,
#     )

#     mocker.patch.object(standings_service, "get", return_value=mock_standings)

#     updated_standings = tabulate_fixture.update_standings(mock_home_team, mock_home_stat)
#     assert updated_standings.points == expected_standings_after_update.points
#     assert updated_standings.score_for == expected_standings_after_update.score_for
#     assert updated_standings.games_played == expected_standings_after_update.games_played

#     team_service.delete(mock_home_team.id)
#     team_service.delete(mock_away_team.id)
#     fixture_team_service.delete(mock_home_stat.id)
#     fixture_team_service.delete(mock_away_stat.id)
#     standings_service.delete(mock_standings.id)


# @pytest.mark.parametrize(["prior_home_points", "expected"], [(0, 3), (1, 4), (6, 9)])
# def test_process_fixture_results(
#     league, schedule, session, standings_service, fixture_team_service, team_service, prior_home_points, expected
# ):
#     mock_home_team = Team(league_id=league.id, name="Mock Home")
#     mock_away_team = Team(league_id=league.id, name="Mock Away")

#     team_service.create(mock_home_team)
#     team_service.create(mock_away_team)

#     fake_fixture = Fixture(
#         home_team_id=mock_home_team.id,
#         away_team_id=mock_away_team.id,
#         schedule_id=uuid4(),
#         status="P",
#         home_goals=3,
#         away_goals=1,
#     )

#     mock_home_stat = FixtureTeam(
#         team_id=mock_home_team.id,
#         fixture_id=fake_fixture.id,
#         is_played=True,
#         points=3,
#         score_for=3,
#         score_against=1,
#         fixture_result="W",
#     )
#     mock_away_stat = FixtureTeam(
#         team_id=mock_away_team.id,
#         fixture_id=fake_fixture.id,
#         is_played=True,
#         points=0,
#         score_for=1,
#         score_against=3,
#         fixture_result="L",
#     )

#     fixture_team_service.create(mock_home_stat)
#     fixture_team_service.create(mock_away_stat)

#     home_standings = Standings(
#         schedule_id=schedule.id,
#         team_id=mock_home_team.id,
#         points=prior_home_points,
#     )

#     away_standings = Standings(
#         schedule_id=schedule.id,
#         team_id=mock_away_team.id,
#     )

#     standings_service.create(home_standings)
#     standings_service.create(away_standings)

#     tabulate_fixture = TabulateFixture(session=session, fixture=fake_fixture)

#     home_standings, away_standings = tabulate_fixture.process_fixture_results()

#     assert home_standings.games_played == 1
#     assert home_standings.games_won == 1
#     assert home_standings.games_drawn == 0
#     assert home_standings.games_lost == 0
#     assert home_standings.score_for == 3
#     assert home_standings.score_against == 1
#     assert home_standings.score_diff == 2
#     assert home_standings.points == expected

#     assert away_standings.games_played == 1
#     assert away_standings.games_won == 0
#     assert away_standings.games_drawn == 0
#     assert away_standings.games_lost == 1
#     assert away_standings.score_for == 1
#     assert away_standings.score_against == 3
#     assert away_standings.score_diff == -2
#     assert away_standings.points == 0

#     team_service.delete(mock_home_team.id)
#     team_service.delete(mock_away_team.id)
#     fixture_team_service.delete(mock_home_stat.id)
#     fixture_team_service.delete(mock_away_stat.id)
#     standings_service.delete(home_standings.id)
#     standings_service.delete(away_standings.id)


# def test_generate_standings_table(schedule, session, standings_service, standings_table):
#     standings_1 = Standings(
#         schedule_id=schedule.id,
#         team_id=standings_table.teams[0].id,
#         points=3,
#         games_played=1,
#         games_won=1,
#         games_drawn=0,
#         games_lost=0,
#         score_for=3,
#         score_against=1,
#         score_diff=-5,
#     )

#     standings_2 = Standings(
#         schedule_id=schedule.id,
#         team_id=standings_table.teams[1].id,
#         points=3,
#         games_played=1,
#         games_won=1,
#         games_drawn=0,
#         games_lost=0,
#         score_for=3,
#         score_against=1,
#         score_diff=2,
#     )

#     standings_3 = Standings(
#         schedule_id=schedule.id,
#         team_id=standings_table.teams[2].id,
#         points=6,
#         games_played=1,
#         games_won=1,
#         games_drawn=0,
#         games_lost=0,
#         score_for=3,
#         score_against=1,
#         score_diff=2,
#     )

#     standings_4 = Standings(
#         schedule_id=schedule.id,
#         team_id=standings_table.teams[3].id,
#         points=12,
#         games_played=1,
#         games_won=1,
#         games_drawn=0,
#         games_lost=0,
#         score_for=3,
#         score_against=1,
#         score_diff=2,
#     )

#     standings_other_schedule = Standings(
#         schedule_id=uuid4(),
#         team_id=standings_table.teams[3].id,
#         points=200,
#         games_played=1,
#         games_won=1,
#         games_drawn=0,
#         games_lost=0,
#         score_for=3,
#         score_against=1,
#         score_diff=2,
#     )

#     standings_other_team = Standings(
#         schedule_id=schedule.id,
#         team_id=uuid4(),
#         points=100,
#         games_played=1,
#         games_won=1,
#         games_drawn=0,
#         games_lost=0,
#         score_for=3,
#         score_against=1,
#         score_diff=2,
#     )

#     clear_table(session, Standings)
#     standings_service.create_many(
#         [standings_1, standings_2, standings_3, standings_4, standings_other_schedule, standings_other_team]
#     )

#     standings = standings_table.generate_standings_table()

#     assert len(standings) == 4
#     assert standings[0].team_id == standings_table.teams[3].id
#     assert standings[0].points == 12
#     assert standings[1].team_id == standings_table.teams[2].id
#     assert standings[2].team_id == standings_table.teams[1].id
#     assert standings[2].points == 3
#     assert standings[2].score_diff == 2
#     assert standings[3].team_id == standings_table.teams[0].id
#     assert standings[3].points == 3
#     assert standings[3].score_diff == -5

#     clear_table(session, Standings)



from .account.address import Address, City, Country, State
from .account.role import Role
from .account.user import User
from .account.user_role import UserRole
from .baseline import (
    Athlete,
    AthleteStats,
    Fixture,
    IndividualMembership,
    League,
    LeagueProperties,
    Manager,
    ManagerMembership,
    Managing,
    Official,
    Officiating,
    Organization,
    Phase,
    Ruleset,
    Season,
    Site,
    Team,
    TeamMembership,
    TeamStats,
)

__all__ = [
    "Address",
    "City",
    "Country",
    "State",
    "Role",
    "User",
    "UserRole",
    "Athlete",
    "AthleteStats",
    "Fixture",
    "IndividualMembership",
    "League",
    "LeagueProperties",
    "Manager",
    "ManagerMembership",
    "Managing",
    "Official",
    "Officiating",
    "Organization",
    "Phase",
    "Season",
    "Ruleset",
    "Site",
    "Team",
    "TeamMembership",
    "TeamStats",
]



from contextlib import asynccontextmanager, contextmanager
from pathlib import Path
from typing import Any, AsyncGenerator, Generator

from attrs import define, field
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session
from svcs import Container, Registry

from leaguemanager.core import get_settings
from leaguemanager.db import async_config, get_async_session, get_session, sync_config, validate_config
from leaguemanager.dependency.loader import DynamicObjectLoader
from leaguemanager.dependency.managers import RepositoryManagement, ServiceManagement
from leaguemanager.services._typing import (
    AsyncRepositoryT,
    AsyncServiceT,
    SQLAlchemyAsyncConfigT,
    SQLAlchemySyncConfigT,
    SyncRepositoryT,
    SyncServiceT,
)

__all__ = ["LeagueManager"]

get_settings.cache_clear()
settings = get_settings()


@define
class LeagueManager:
    """Registry for managing services.

    TODO: Serve up async repos/services

    If no `Registry` is provided, one will be created. Keep in mind that there should only
    be one registry per application.

    Services are kept in an svcs `Container` and are provided as needed. This includes a
    database session, League Manager "repositories" and "services" (which themselves provide
    common database operations), as well as Advanced Alchemy database configuration objects.

    Attributes:
        service_registry (Registry | None): An `svcs` Registry for managing services.
        loader (DynamicObjectLoader): A DynamicObjectLoader for loading specific objects.
        local_base_dir (Path): The local base directory. Uses `settings.APP_DIR` by default.
        local_root_dir (Path): The local root directory. Uses `settings.APP_ROOT` by default.
        aa_config_dir (Path): The Advanced Alchemy configuration directory.
        get_session (Generator[Session, Any, None]): A generator for a database session.
        get_async_session (AsyncGenerator[AsyncSession, Any]): A generator for an async database session.
        manager_sync_repos (list[type[SyncRepositoryT]]): List of repositories for sync database operations.
        manager_async_repos (list[type[AsyncRepositoryT]]): List of repositories for async database operations.
        manager_sync_services (list[type[SyncServiceT]]): List of services for sync database operations.
        manager_async_services (list[type[AsyncServiceT]]): List of services for async database operations.

    Example:
        >>> registry = LeagueManager()
        >>> season_service = registry.provide_db_service(SeasonSyncService)
        >>> team_service = registry.provide_db_service(TeamSyncService)
        >>>
        >>> season_service.list()  #  List all seasons
        >>> team_service.count()  #  Count number of teams

    """

    service_registry: Registry | None = field(default=None)

    loader: DynamicObjectLoader = field(default=DynamicObjectLoader())

    # Might make sense to make these private or keep them in post_init
    # Better to control these through the environment variables
    local_base_dir: Path = field(default=settings.APP_DIR)
    local_root_dir: Path = field(default=settings.APP_ROOT)
    aa_config_dir: Path | None = field(default=settings.ADVANCED_ALCHEMY_CONFIG_DIR)

    get_session: Generator[Session, Any, None] = field(default=get_session)
    get_async_session: AsyncGenerator[AsyncSession, Any] = field(default=get_async_session)

    sync_repos: list[type[SyncRepositoryT]] = field(init=False)
    async_repos: list[type[AsyncRepositoryT]] = field(init=False)
    sync_services: list[type[SyncServiceT]] = field(init=False)
    async_services: list[type[AsyncServiceT]] = field(init=False)

    # Useful for integrating with other apps/frameworks
    _SVCS_KEY_REGISTRY: str = field(default="league_manager_registry")
    _SVCS_KEY_CONTAINER: str = field(default="league_manager")

    def __attrs_post_init__(self):
        if not self.service_registry:
            self.service_registry = Registry()

        # Get Advanced Alchemy objects
        self.sync_repos = self.loader.get_repositories()
        self.async_repos = self.loader.get_repositories(is_async=True)
        self.sync_services = self.loader.get_services()
        self.async_services = self.loader.get_services(is_async=True)

        if not settings.APP_DIR.name == settings.DEFAULT_MODULE_NAME and not settings.SEARCH_APP_MODULES:
            self.loader = self.loader.local_app()
            self.sync_repos += self.loader.get_repositories()
            self.async_repos += self.loader.get_repositories(is_async=True)
            self.sync_services += self.loader.get_services()
            self.async_services += self.loader.get_services(is_async=True)

        if not self.aa_config_dir:
            print("No ENV VAR for Advanced Alchemy config found. Using default config.")
            _sync_config = [sync_config]
            _async_config = [async_config]
        else:
            print(f"Loading Advanced Alchemy configurations from {self.aa_config_dir}")
            self.loader = self.loader.local_app(search_dir=self.loader.app_dir / self.aa_config_dir)
            _sync_config = self.loader.get_configs()
            _async_config = self.loader.get_configs(is_async=True)
            if not _sync_config and not _async_config:
                raise ValueError(
                    f"Cannot find any Advanced Alchemy configurations in {self.aa_config_dir}. "
                    "You may need to wrap the configuration within a function."
                )

        # Register objects
        for repo_type in self.sync_repos:
            self.register_db_repository(repository_type=repo_type)
        for service_type in self.sync_services:
            self.register_db_service(service_type=service_type)

        for _config in _sync_config:
            _config = validate_config(_config)
            self.registry.register_value(SQLAlchemySyncConfigT, _config)
        for _config in _async_config:
            _config = validate_config(_config, is_async=True)
            self.registry.register_value(SQLAlchemyAsyncConfigT, _config)

    @property
    def registry(self) -> Registry:
        return self.service_registry

    @contextmanager
    def sync_session_container(self) -> Generator[Container, None, None]:
        """Create a container for a sync database session."""
        self.registry.register_factory(Session, self.get_session)
        with Container(self.registry) as container:
            yield container

    @asynccontextmanager
    async def async_session_container(self) -> AsyncGenerator[Container, None]:
        """Create a container for an async database session."""
        self.registry.register_factory(AsyncSession, self.get_async_session)
        async with Container(self.registry) as container:
            yield container

    def register_db_repository(self, repository_type: type[SyncRepositoryT]) -> None:
        """Register a League Manager repository based on its type."""
        _repository = RepositoryManagement(repository_type=repository_type, db_session=self.provide_db_session)
        self.registry.register_value(repository_type, _repository.get_repository)

    def register_db_service(self, service_type: type[SyncServiceT]) -> None:
        """Register a League Manager service based on its type."""
        _service = ServiceManagement(service_type=service_type, db_session=self.provide_db_session)
        self.registry.register_value(service_type, next(_service.get_service))

    # Retrieve and provide dependencies

    @property
    def provide_db_session(self) -> Session:
        """Provide a sync database session."""
        with self.sync_session_container() as container:
            return container.get(Session)

    @property
    def provide_async_db_session(self) -> AsyncSession:
        """Provide an async database session."""
        with self.async_session_container() as container:
            return container.get(AsyncSession)

    @property
    def provide_sync_config(self) -> SQLAlchemySyncConfigT:
        return Container(self.registry).get(SQLAlchemySyncConfigT)

    @property
    def provide_async_config(self) -> SQLAlchemyAsyncConfigT:
        return Container(self.registry).get(SQLAlchemyAsyncConfigT)

    def provide_db_repository(self, repository_type: type[SyncRepositoryT]) -> type[SyncRepositoryT]:
        """Provide a League Manager repository based on its type."""
        return Container(self.registry).get(repository_type)

    def provide_db_service(self, service_type: type[SyncServiceT]) -> type[SyncServiceT]:
        """Provide a League Manager service based on its type."""
        return Container(self.registry).get(service_type)



from .athlete import Athlete
from .individual_membership import IndividualMembership
from .manager import Manager
from .manager_membership import ManagerMembership
from .official import Official
from .team import Team
from .team_membership import TeamMembership

__all__ = [
    "Athlete",
    "IndividualMembership",
    "Manager",
    "ManagerMembership",
    "Official",
    "Team",
    "TeamMembership",
]



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import AthleteStats
from leaguemanager.repository import AthleteStatsSyncRepository
from leaguemanager.repository._async import AthleteStatsAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["AthleteStatsService", "AthleteStatsAsyncService"]


class AthleteStatsService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = AthleteStatsSyncRepository


class AthleteStatsAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = AthleteStatsAsyncRepository



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager import models as m
from leaguemanager.repository import SeasonSyncRepository
from leaguemanager.repository._async import SeasonAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["SeasonService", "SeasonAsyncService"]


class SeasonService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = SeasonSyncRepository

    def all_teams(self, season_id: UUID, active: bool = True) -> list[m.Team]:
        """Get all active teams for a given season.

        Selects all teams based off their associaton to a TeamMembership object, which
        in turn is associated with a Season object. The Team must be active."""

        season = self.get(season_id)
        memberships = season.team_memberships
        teams = []
        if not memberships:
            return []
        for _memb in memberships:
            if active:
                if not _memb.team.active:
                    continue
            teams.append(_memb.team)
        return teams

    def all_athletes(self, season_id: UUID, incl_inactive_team: bool = False) -> list[m.Athlete]:
        """Get all players for a given season.

        Selects all players based off their associaton to a IndividualMembership object.
        """

        season = self.get(season_id)
        team_memberships = season.team_memberships
        athletes = []
        for _member in team_memberships:
            team = _member.team
            if team is None:
                continue
            if not incl_inactive_team and not team.active:
                continue
            individual_members = team.individual_memberships
            if individual_members is None:
                continue
            for _indiv in individual_members:
                athletes.append(_indiv.athlete)
        return athletes


class SeasonAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = SeasonAsyncRepository

    async def all_teams(self, season_id: UUID, active: bool = True) -> list[m.Team]:
        """Get all active teams for a given season.

        Selects all teams based off their associaton to a TeamMembership object, which
        in turn is associated with a Season object. The Team must be active."""

        season = await self.get(season_id)
        memberships = season.team_memberships
        teams = []
        if not memberships:
            return []
        for _memb in memberships:
            if active:
                if not _memb.team.active:
                    continue
            teams.append(_memb.team)
        return teams

    async def all_athletes(self, season_id: UUID, incl_inactive_team: bool = False) -> list[m.Athlete]:
        """Get all players for a given season.

        Selects all players based off their associaton to a IndividualMembership object.
        """

        season = await self.get(season_id)
        team_memberships = season.team_memberships
        athletes = []
        for _member in team_memberships:
            team = _member.team
            if team is None:
                continue
            if not incl_inactive_team and not team.active:
                continue
            individual_members = team.individual_memberships
            if individual_members is None:
                continue
            for _indiv in individual_members:
                athletes.append(_indiv.athlete)
        return athletes



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import CompetitionBase, mapper, metadata

if TYPE_CHECKING:
    from leaguemanager.models.baseline import TeamMembership

    from .league import League


@define(slots=False)
class Season(CompetitionBase):
    """Defines a competition type with a beginning and end, and usually has winner at the end.

    This is usually a child of a League object, with start and end dates. While the League
    is perpetual (recurring), a Season object tend to be finite (i.e., Spring Volleball 2025,
    Winter/Spring '26).

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        league_id (UUID): ForeignKey to League.
        name (str): Descriptive name. (i.e, Fall Coed Soccer 2025, Summer Little League 2026)
        description (str): Optional description of the season details.
        # competition_type (str): Desginate the type of competition. Inherited from CompetitionBase. (NOT MAPPED)
        active (bool): Defaults to True. Can be used to deactivate this Season.
        projected_start_date (datetime): When the Season is projected to start.
        projected_end_date (datetime): When the Season is project to end.
        actual_start_date (datetime): When the Season actually begins.
        actual_end_date (datetime): When the Season actually ends.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    league_id: UUID | None = field(default=None)
    name: str | None = field(default=None, validator=validators.max_len(80))
    description: str | None = field(default=None, validator=validators.optional(validators.max_len(120)))
    active: bool = field(default=True)
    projected_start_date: datetime | None = field(
        default=None, converter=lambda d: datetime.fromisoformat(d) if isinstance(d, str) else d
    )
    projected_end_date: datetime | None = field(default=None)
    actual_start_date: datetime | None = field(default=None)
    actual_end_date: datetime | None = field(default=None)

    # team_memberships: list[TeamMembership] | None = None


# SQLAlchemy Imperative Mappings

season = Table(
    "season",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("league_id", SA_UUID, ForeignKey("league.id"), nullable=True),
    Column(
        "name",
        String(80),
        nullable=False,
        unique=True,
    ),
    Column("description", String(120), nullable=True),
    Column("active", Boolean, default=True),
    Column("projected_start_date", DateTime(), nullable=True, default=None),
    Column("projected_end_date", DateTime(), nullable=True, default=None),
    Column("actual_start_date", DateTime(), nullable=True, default=None),
    Column("actual_end_date", DateTime(), nullable=True, default=None),
    # Column("cost", Float, nullable=True, default=None),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    Season,
    season,
    properties={
        "site": relationship("Site", back_populates="season"),
        "league": relationship("League", back_populates="seasons"),
        "phases": relationship("Phase", back_populates="season"),
        "fixtures": relationship("Fixture", back_populates="season"),
        "team_memberships": relationship("TeamMembership", back_populates="season"),
        "athlete_stats": relationship("AthleteStats", back_populates="season"),
        "team_stats": relationship("TeamStats", back_populates="season"),
        "ruleset": relationship("Ruleset", back_populates="season", uselist=False, cascade="all, delete-orphan"),
    },
)



import importlib
import json
import os
from typing import Annotated, Optional

import typer
from attrs import asdict, define, field
from httpx import Client, Headers, Response
from rich import print


@define
class Release:
    tag_name: str
    name: str
    body: str
    draft: bool = False
    prerelease: bool = False
    hide_archive_links: bool = True


@define
class ReleaseBody:
    text: str = "# Highlights\r\n"

    def add_line(self, line: str):
        self.text += f"{line}\n"


@define
class CBAPI:
    cb_api_key: str = field(default=(os.getenv("CODEBERG_API_KEY")))
    user: str = field(default=(os.getenv("CODEBERG_REPO_USER")))
    repo: str = field(default="leaguemanager")
    base_url: str = field()
    headers: Headers = field()
    client: Client = field()

    @base_url.default
    def _default_base_url(self) -> str:
        return f"https://codeberg.org/api/v1/repos/{self.user}/{self.repo}"

    @headers.default
    def _default_headers(self) -> Headers:
        headers = Headers({"Authorization": f"token {self.cb_api_key}", "Content-Type": "application/json"})
        return headers

    @client.default
    def _default_client(self) -> Client:
        return Client(base_url=self.base_url, headers=self.headers)

    @property
    def current_version(self) -> str:
        return importlib.import_module("leaguemanager").__version__

    def get_relases(self) -> Response:
        return self.client.get("releases")

    def get_latest_release(self) -> Response:
        return self.client.get("releases/latest")

    def create_release(self, data: dict) -> Response:
        # test_data = {
        #     "body": "Add CLI for auto release.",
        #     "draft": True,
        #     "hide_archive_links": True,
        #     "name": "v0.5.4",
        #     "prerelease": False,
        #     "tag_name": "v0.5.4",
        #     "target_commitish": "string",
        # }
        return self.client.post("releases", json=data)


app = typer.Typer(no_args_is_help=True)
cb = CBAPI()


def _release_body(create: bool) -> ReleaseBody:
    if not create:
        return
    body = ReleaseBody()
    current_version = cb.current_version

    line = typer.prompt("Input release line: ")
    body.add_line(line)
    more = typer.confirm("Add more release lines?")
    while more:
        line = typer.prompt("Input release line: ")
        body.add_line(line)
        more = typer.confirm("Add more release lines?")
    release = Release(
        tag_name=f"v{current_version}",
        name=f"v{current_version}",
        body=body.text,
    )
    typer.confirm("Publish to Codeberg?")
    return asdict(release)


@app.callback(no_args_is_help=True)
def main():
    return


@app.command()
def releases(
    create: bool = typer.Option(
        False, "--create", "-c", is_flag=True, callback=_release_body, help="Create a release."
    ),
    latest: bool = typer.Option(False, "--latest", "-l", is_flag=True, help="Get the latest release."),
) -> None:
    """Release commands."""
    if latest:
        latest_release = cb.get_latest_release().json()
        print(json.dumps(latest_release, indent=4))
        return
    if create:
        _ = cb.create_release(data=create)
        print(f"Created Release v{cb.current_version} in Codeberg.")
        return

    releases = cb.get_relases().json()
    print(json.dumps(releases, indent=4))
    return


if __name__ == "__main__":
    app(prog_name="cbapi")



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Officiating
from leaguemanager.repository import OfficiatingSyncRepository
from leaguemanager.repository._async import OfficiatingAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["OfficiatingService", "OfficiatingAsyncService"]


class OfficiatingService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = OfficiatingSyncRepository


class OfficiatingAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = OfficiatingAsyncRepository



from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table, UniqueConstraint
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import UUIDBase, mapper, metadata
from leaguemanager.models.enums import Category, Division, MatchDay

if TYPE_CHECKING:
    from leaguemanager.models.baseline.competition import Site


@define(slots=False)
class Address(UUIDBase):
    """Defines the street address of another object"""

    site_id: UUID | None = field(default=None)
    city_id: UUID | None = field(default=None)
    street: str | None = field(default=None, validator=validators.optional(validators.max_len(120)))
    postal_code: str | None = field(default=None, validator=validators.optional(validators.max_len(20)))


@define(slots=False)
class City(UUIDBase):
    """Defines a city with a name and an optional country code."""

    state_id: UUID | None = field(default=None)
    city_name: str | None = field(default=None, validator=validators.max_len(80))


@define(slots=False)
class State(UUIDBase):
    """Defines a state with a name and an optional country code."""

    country_id: UUID | None = field(default=None)
    state_name: str | None = field(default=None, validator=validators.max_len(80))
    state_code: str | None = field(default=None, validator=validators.optional(validators.max_len(3)))


@define(slots=False)
class Country(UUIDBase):
    """Defines a country with a name and an optional country code."""

    country_name: str | None = field(default=None, validator=validators.max_len(80))
    country_code: str | None = field(default=None, validator=validators.optional(validators.max_len(3)))


# SQLAlchemy Imperative Mappings
address = Table(
    "address",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("site_id", SA_UUID, ForeignKey("site.id"), nullable=True),
    Column("city_id", SA_UUID, ForeignKey("city.id"), nullable=True),
    Column("street", String(120), nullable=True),
    Column("postal_code", String(20), nullable=True),
)

city = Table(
    "city",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("state_id", SA_UUID, ForeignKey("state.id"), nullable=True),
    Column("city_name", String(80), nullable=False),
)
state = Table(
    "state",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("country_id", SA_UUID, ForeignKey("country.id"), nullable=True),
    Column("state_name", String(80), nullable=False),
    Column("state_code", String(3), nullable=True),
)
country = Table(
    "country",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("country_name", String(80), nullable=False),
    Column("country_code", String(3), nullable=True),
)

# ORM Relationships
mapper.map_imperatively(
    Address,
    address,
    properties={
        "site": relationship("Site", back_populates="address", uselist=False),
        "city": relationship("City", back_populates="addresses"),
    },
)

mapper.map_imperatively(
    City,
    city,
    properties={
        "state": relationship("State", back_populates="cities"),
        "addresses": relationship("Address", back_populates="city", uselist=True),
    },
)

mapper.map_imperatively(
    State,
    state,
    properties={
        "country": relationship("Country", back_populates="states"),
        "cities": relationship("City", back_populates="state", uselist=True),
    },
)

mapper.map_imperatively(
    Country,
    country,
    properties={
        "states": relationship("State", back_populates="country", uselist=True),
    },
)



from advanced_alchemy.repository import (
    SQLAlchemyAsyncRepository,
    SQLAlchemyAsyncSlugRepository,
)

from leaguemanager import models

__all__ = [
    "AddressAsyncRepository",
    "CityAsyncRepository",
    "CountryAsyncRepository",
    "StateAsyncRepository",
    "RoleAsyncRepository",
    "UserAsyncRepository",
    "UserRoleAsyncRepository",
    "AthleteAsyncRepository",
    "AthleteStatsAsyncRepository",
    "FixtureAsyncRepository",
    "IndividualMembershipAsyncRepository",
    "LeagueAsyncRepository",
    "LeaguePropertiesAsyncRepository",
    "ManagerAsyncRepository",
    "ManagerMembershipAsyncRepository",
    "ManagingAsyncRepository",
    "OfficialAsyncRepository",
    "OfficiatingAsyncRepository",
    "OrganizationAsyncRepository",
    "PhaseAsyncRepository",
    "RulesetAsyncRepository",
    "SeasonAsyncRepository",
    "SiteAsyncRepository",
    "TeamAsyncRepository",
    "TeamMembershipAsyncRepository",
    "TeamStatsAsyncRepository",
]


class AddressAsyncRepository(SQLAlchemyAsyncRepository[models.Address]):
    """Address repository."""

    model_type = models.Address


class CityAsyncRepository(SQLAlchemyAsyncRepository[models.City]):
    """City repository."""

    model_type = models.City


class CountryAsyncRepository(SQLAlchemyAsyncRepository[models.Country]):
    """Country repository."""

    model_type = models.Country


class StateAsyncRepository(SQLAlchemyAsyncRepository[models.State]):
    """State repository."""

    model_type = models.Country


class RoleAsyncRepository(SQLAlchemyAsyncSlugRepository[models.Role]):
    """Role repository."""

    model_type = models.Role


class UserAsyncRepository(SQLAlchemyAsyncRepository[models.User]):
    """User repository."""

    model_type = models.User


class UserRoleAsyncRepository(SQLAlchemyAsyncRepository[models.UserRole]):
    """UserRole repository."""

    model_type = models.UserRole


class AthleteAsyncRepository(SQLAlchemyAsyncRepository[models.Athlete]):
    """Athlete repository."""

    model_type = models.Athlete


class AthleteStatsAsyncRepository(SQLAlchemyAsyncRepository[models.AthleteStats]):
    """AthleteStats repository."""

    model_type = models.AthleteStats


class FixtureAsyncRepository(SQLAlchemyAsyncRepository[models.Fixture]):
    """Fixture repository."""

    model_type = models.Fixture


class IndividualMembershipAsyncRepository(SQLAlchemyAsyncRepository[models.IndividualMembership]):
    """IndividualMembership repository."""

    model_type = models.IndividualMembership


class LeagueAsyncRepository(SQLAlchemyAsyncSlugRepository[models.League]):
    """League repository."""

    model_type = models.League


class LeaguePropertiesAsyncRepository(SQLAlchemyAsyncSlugRepository[models.LeagueProperties]):
    """LeagueProperties repository."""

    model_type = models.LeagueProperties


class ManagerAsyncRepository(SQLAlchemyAsyncRepository[models.Manager]):
    """Manager repository."""

    model_type = models.Manager


class ManagerMembershipAsyncRepository(SQLAlchemyAsyncRepository[models.ManagerMembership]):
    """ManagerMembership repository."""

    model_type = models.ManagerMembership


class ManagingAsyncRepository(SQLAlchemyAsyncRepository[models.Managing]):
    """Managing repository."""

    model_type = models.Managing


class OfficialAsyncRepository(SQLAlchemyAsyncRepository[models.Official]):
    """Official repository."""

    model_type = models.Official


class OfficiatingAsyncRepository(SQLAlchemyAsyncRepository[models.Officiating]):
    """Officiating repository."""

    model_type = models.Officiating


class OrganizationAsyncRepository(SQLAlchemyAsyncSlugRepository[models.Organization]):
    """Organization repository."""

    model_type = models.Organization


class PhaseAsyncRepository(SQLAlchemyAsyncRepository[models.Phase]):
    """Phase repository."""

    model_type = models.Phase


class RulesetAsyncRepository(SQLAlchemyAsyncSlugRepository[models.Ruleset]):
    """Ruleset repository."""

    model_type = models.Ruleset


class SeasonAsyncRepository(SQLAlchemyAsyncSlugRepository[models.Season]):
    """Season repository."""

    model_type = models.Season


class SiteAsyncRepository(SQLAlchemyAsyncSlugRepository[models.Site]):
    """Site repository."""

    model_type = models.Site


class TeamAsyncRepository(SQLAlchemyAsyncSlugRepository[models.Team]):
    """Team repository."""

    model_type = models.Team


class TeamMembershipAsyncRepository(SQLAlchemyAsyncSlugRepository[models.TeamMembership]):
    """TeamMembership repository."""

    model_type = models.TeamMembership


class TeamStatsAsyncRepository(SQLAlchemyAsyncSlugRepository[models.TeamStats]):
    """TeamStats repository."""

    model_type = models.TeamStats



import os
from functools import lru_cache
from importlib.util import find_spec
from pathlib import Path

from attrs import define, field


@define
class Settings:
    # League Manager app info
    DEFAULT_MODULE_NAME: str = "leaguemanager"
    LEAGUE_MANAGER_APP_DIR: Path = field()
    LEAGUE_MANAGER_ROOT_DIR: Path = field()

    @LEAGUE_MANAGER_APP_DIR.default
    def _default_module_path(self):
        return self._module_to_os_path(self.DEFAULT_MODULE_NAME)

    @LEAGUE_MANAGER_ROOT_DIR.default
    def _default_root_path(self):
        return self.LEAGUE_MANAGER_APP_DIR.parent.resolve()

    DATE_FORMAT: str = field(default="%Y-%m-%d %H:%M:%S")
    SYNTH_DATA_DIR: Path = field(init=False)

    SEARCH_APP_MODULES: bool = field(default=False)

    def __attrs_post_init__(self):
        self.SYNTH_DATA_DIR: Path = Path(self.LEAGUE_MANAGER_APP_DIR) / "db/synthetic_data"

        if _search_app := os.getenv("SEARCH_APP_MODULES"):
            self.SEARCH_APP_MODULES = bool(_search_app)

    @property
    def APP_DIR(self):
        if _app_dir := os.getenv("APP_DIR"):
            try:
                app_dir = Path(_app_dir)
                if app_dir.is_dir():
                    return app_dir
            except ValueError:
                print(f"Environment Variable refers to invalid directory: {_app_dir}")
        return Path.cwd()

    @property
    def APP_ROOT(self):
        if _app_root := os.getenv("APP_ROOT"):
            try:
                app_root = Path(_app_root)
                if app_root.is_dir():
                    return app_root
            except ValueError:
                print(f"Environment Variable refers to invalid directory: {_app_root}")
        return Path.cwd()

    @property
    def ADVANCED_ALCHEMY_CONFIG_DIR(self):
        if self.APP_DIR.name == self.DEFAULT_MODULE_NAME:
            return None
        if _aa_config_module := os.getenv("ADVANCED_ALCHEMY_CONFIG_DIR"):
            _aa_config_dir = self.APP_DIR / _aa_config_module
            if not Path(_aa_config_dir).is_dir():
                raise ValueError(f"Environment Variable refers to invalid directory: {_aa_config_dir}")
            return _aa_config_dir
        else:
            return None

    @property
    def MIGRATION_PATH(self):
        if "MIGRATION_PATH" in os.environ:
            migration_path = Path(os.getenv("MIGRATION_PATH"))
            return Path(self.APP_DIR / migration_path, exist_ok=True)
        else:
            return self.APP_DIR / "migrations"

    @property
    def MIGRATION_CONFIG(self):
        if "MIGRATION_CONFIG" in os.environ:
            return Path(os.getenv("MIGRATION_CONFIG"))
        else:
            return self.APP_DIR

    @property
    def ALEMBIC_TEMPLATE_PATH(self):
        if "ALEMBIC_TEMPLATE_PATH" in os.environ:
            return Path(os.getenv("ALEMBIC_TEMPLATE_PATH"))
        else:
            return Path(self.LEAGUE_MANAGER_APP_DIR) / "db/alembic_templates"

    @property
    def SQLITE_DATA_DIRECTORY(self):
        if "SQLITE_DATA_DIRECTORY" in os.environ:
            return os.getenv("SQLITE_DATA_DIRECTORY")
        return self.APP_DIR / "data_league_db"

    def _module_to_os_path(self, module_name: str) -> Path:
        """Get the string path of the module."""
        spec = find_spec(module_name)
        if not spec:
            raise ValueError(f"Couldn't find path for {module_name}")
        return Path(spec.origin).parent.resolve()


@lru_cache
def get_settings() -> Settings:
    return Settings()



from __future__ import annotations

from typing import Iterator

from attrs import define, field
from sqlalchemy import select
from sqlalchemy.orm import Session

from leaguemanager.services._typing import AsyncRepositoryT, AsyncServiceT, ModelT, SyncRepositoryT, SyncServiceT

__all__ = ["ServiceManagement", "RepositoryManagement"]


@define
class ServiceManagement:
    """Manages a SQLAlchemySyncRepositoryService[ModelT] class.

    TODO: Async support

    Given the `service_type` and `db_session`, as well as a db_session, it will hold then provide
    the applicable service (for the corresponding service type.). The `get_service` property will
    return the appropriate service for the given `service_type` and `db_session`.

    Attributes:
        service_type (type[SyncServiceT] | type[AsyncServiceT]): Service type to manage.
        model_type (type[ModelT]): Model type for the given `service_type`.
        db_session (Session | None): Database session to use for the service.

    Example:
      >>> _service = ServiceManagement(
      ...     service_type=SeasonSyncService, model_type=Season, db_session=session
      ... )
      >>> _service.get_service
    """

    service_type: type[SyncServiceT] = field()
    model_type: type[ModelT] = field(init=False)
    db_session: Session = field(default=None)

    def __attrs_post_init__(self):
        self.model_type = self.service_type.repository_type.model_type

    @property
    def get_service(self) -> Iterator[ServiceManagement.service_type]:
        with self.service_type.new(session=self.db_session, statement=select(self.model_type)) as service:
            yield service


@define
class RepositoryManagement:
    """Manages a SQLAlchemySyncRepository class.

    TODO: Async support

    Given the `repository_type` as well as a db_session, it will hold then provide
    the applicable repository (for the corresponding repository type.). The `get_repository` property will
    return the appropriate repository for the given `repository_type` and `db_session`.

    Attributes:
        repository_type (type[SyncRepositoryT] | type[AsyncRepositoryT]): Repository type to manage.
        db_session (Session | None): Database session to use for the repository.

    Example:
      >>> _repository = RepositoryManagement(
      ...     repository_type=SeasonSyncRepository, db_session=session
      ... )
      >>> _repository.get_repository
    """

    repository_type: type[SyncRepositoryT] = field()
    db_session: Session = field(default=None)

    @property
    def get_repository(self) -> Iterator[RepositoryManagement.repository_type]:
        return self.repository_type(session=self.db_session, statement=select(self.repository_type.model_type))



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Role
from leaguemanager.repository import RoleSyncRepository
from leaguemanager.repository._async import RoleAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["RoleSyncService", "RoleAsyncService"]


class RoleSyncService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = RoleSyncRepository


class RoleAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = RoleAsyncRepository



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import IndividualMembership
from leaguemanager.repository import IndividualMembershipSyncRepository
from leaguemanager.repository._async import IndividualMembershipAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["IndividualMembershipService", "IndividualMembershipAsyncService"]


class IndividualMembershipService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = IndividualMembershipSyncRepository


class IndividualMembershipAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = IndividualMembershipAsyncRepository



"""Entry point for CLI application."""

from leaguemanager import __app_name__, cli


def main() -> None:
    """Entrypoint for CLI application."""
    cli.app(prog_name=__app_name__)


if __name__ == "__main__":
    main()



import os

from leaguemanager.core import get_settings

get_settings.cache_clear()
settings = get_settings()

SYNC_DATABASE_URI = os.getenv("SYNC_DATABASE_URI", None)
ASYNC_DATABASE_URI = os.getenv("ASYNC_DATABASE_URI", None)

ECHO: bool = os.getenv("ECHO", False)


def _is_sqlite(uri: str | None) -> bool:
    if uri is None:
        return False
    settings.SQLITE_DATA_DIRECTORY.mkdir(parents=True, exist_ok=True)
    return "sqlite" in uri.lower()


def _create_str_uri(is_async: bool = False) -> str:
    uri = SYNC_DATABASE_URI if not is_async else ASYNC_DATABASE_URI
    if not _is_sqlite(uri):
        try:
            settings.SQLITE_DATA_DIRECTORY.mkdir(parents=True, exist_ok=True)
            if is_async:
                return "sqlite+aiosqlite:///data_league_db/lmgr_data.db"
            return "sqlite:///data_league_db/lmgr_data.db"
        except ValueError as e:
            raise NotImplementedError(f"Database implementation needed. {e}") from ValueError
    if not uri:
        raise ValueError("DATABASE_URI is not set")
    return uri


def uri(is_async: bool = False) -> str:
    if is_async:
        return _create_str_uri(is_async=True)
    return _create_str_uri()


db_args = {"echo": ECHO, "connect_args": {"check_same_thread": False}}



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import TeamMembership
from leaguemanager.repository import TeamMembershipSyncRepository
from leaguemanager.repository._async import TeamMembershipAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["TeamMembershipService", "TeamMembershipAsyncService"]


class TeamMembershipService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = TeamMembershipSyncRepository


class TeamMembershipAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = TeamMembershipAsyncRepository



import pytest
from sqlalchemy import delete
from sqlalchemy.ext.asyncio import AsyncSession

from leaguemanager import models as m
from leaguemanager import services as s
from leaguemanager.core.toolbox import clear_table

org_data = {
    "name": "Test Organization",
}

league_data = {
    "name": "Test League",
}


season_data = {
    "name": "Season Test",
    "description": "Burndt",
    "projected_start_date": "2022-01-01",
}


team_data = [
    {"name": "Loro"},
    {"name": "June"},
    {"name": "Tripoli"},
    {"name": "Tres", "active": False},
]


@pytest.fixture(scope="module")
def org(org_service):
    return org_service.create(
        {
            "name": "Test Organization",
        },
        auto_commit=True,
    )


@pytest.fixture(scope="module")
def league():
    return m.League(**league_data)


@pytest.fixture(scope="module")
def properties(league_properties_service):
    return league_properties_service.create(
        {"sport": "Tiddlywinks", "category": "COED", "division": "A"},
        auto_commit=True,
    )


@pytest.fixture(scope="module")
def ruleset(ruleset_service):
    return ruleset_service.create(
        {
            "number_of_games": 10,
            "sun_fixtures": True,
        },
        auto_commit=True,
    )


@pytest.fixture(scope="module")
def season():
    return m.Season(**season_data)


@pytest.fixture(scope="module")
def teams():
    return [m.Team(**data) for data in team_data]


def test_create_many_leagues(session, league_service) -> None:
    new_leagues_data = [
        {
            "name": "Test Create League",
            "description": "Test Description",
            "sport": "Football",
        },
        {
            "name": "Test Create League 2",
            "description": "Test Description 2",
            "sport": "Football",
        },
    ]

    _ = league_service.create_many(new_leagues_data)

    assert league_service.count() == 2
    clear_table(session, m.League)


def test_delete_league(league_service) -> None:
    new_league = m.League(name="Test Delete League 2", description="Hi")
    test_league = league_service.create(new_league)
    league_service.delete(test_league.id)
    assert league_service.get_one_or_none(name="Test Delete League 2") is None


def test_org_with_league(session, league_service, season_service, org) -> None:
    new_league = m.League(name="Test Create League", description="boo", organization_id=org.id)
    temp_league = league_service.create(new_league, auto_commit=True)

    assert temp_league.name == "Test Create League"
    assert temp_league.organization_id == org.id

    clear_table(session, m.League)


def test_league_with_properties(session, league_service, league_properties_service, league, properties) -> None:
    new_league = league_service.create(league, auto_commit=True)
    properties.league_id = new_league.id
    league_properties_service.update(properties, auto_commit=True)

    _properties = new_league.league_properties

    assert new_league.name == "Test League"
    assert _properties.sport == "Tiddlywinks"

    clear_table(session, m.League)
    clear_table(session, m.LeagueProperties)


def test_league_with_season(session, league_service, season_service, org, league, season) -> None:
    new_league = league_service.create(league, auto_commit=True)
    new_league.organization_id = org.id
    new_season = season_service.create(season, auto_commit=True)
    new_season.league_id = new_league.id
    new_league.seasons.append(new_season)

    assert new_league.name == "Test League"
    assert new_season.name == "Season Test"

    clear_table(session, m.League)
    clear_table(session, m.Season)


def test_append_league_to_season(session, season_service, league_service) -> None:
    league_data = {
        "name": "Songs 2",
        "description": "I do a sport",
        "sport": "Music",
    }
    season_data = {
        "name": "SZNZ 2",
        "description": "I Want A Dog",
        "projected_start_date": "2025-01-01",
    }

    league = league_service.create(data=league_data, auto_commit=True)
    season = season_service.create(data=season_data, auto_commit=True)

    league.seasons.append(season)

    assert league_service.get_one_or_none(name="Songs 2").seasons[0].name == "SZNZ 2"
    assert season_service.get_one_or_none(name="SZNZ 2").league.name == "Songs 2"

    clear_table(session, m.League)
    clear_table(session, m.Season)


def test_season_with_ruleset(session, season_service, league_service, org, league, season, ruleset) -> None:
    new_league = league_service.create(league, auto_commit=True)
    new_league.organization_id = org.id
    new_season = season_service.create(season, auto_commit=True)
    new_season.league_id = new_league.id

    new_season.ruleset = ruleset

    assert new_season.ruleset.number_of_games == 10
    assert new_season.ruleset.sun_fixtures is True

    clear_table(session, m.League)
    clear_table(session, m.Season)
    clear_table(session, m.Ruleset)



from attrs import define, field, validators
from sqlalchemy import UUID as _UUID
from sqlalchemy import Column, String, Table
from sqlalchemy.orm import relationship

from leaguemanager.models.base import UUIDBase, mapper, metadata


@define(slots=False)
class Role(UUIDBase):
    """A role defining permissions in the system."""

    name: str | None = field(default=None, validator=validators.max_len(12))
    description: str | None = field(default=None, validator=validators.optional(validators.max_len(80)))


# SQLAlchemy Imperative Mapping

role = Table(
    "role",
    metadata,
    Column("id", _UUID, primary_key=True),
    Column("name", String(12), nullable=False),
    Column("description", String(80), nullable=True),
)

# ORM Relationships

mapper.map_imperatively(Role, role, properties={"user_roles": relationship("UserRole", back_populates="role")})



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Athlete
from leaguemanager.repository import AthleteSyncRepository
from leaguemanager.repository._async import AthleteAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["AthleteService", "AthleteAsyncService"]


class AthleteService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = AthleteSyncRepository


class AthleteAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = AthleteAsyncRepository



from __future__ import annotations

from datetime import UTC, datetime
from uuid import UUID

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Column, DateTime, ForeignKey, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import CompetitionPhaseBase, mapper, metadata

if TYPE_CHECKING:
    from .league import League


@define(slots=False)
class Phase(CompetitionPhaseBase):
    """Defines a 'phase' of a competition.

    A phase represents a distinct phase within a Season. For example, while a given
    Competition could be completed in one or several Fixtures, a Phase corresponds to
    a series of Fixtures, such as a 'Matchday' or 'Match Week'. Alternatively, a
    Phase could correspond to a phase within a competition, such as quarterfinals,
    semifinals, or final match.

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        season_id (UUID): ForeignKey to Season.
        label (str): Unique label describing the Phase (i.e., Matchday 3, Gameweek 12, Semi-Final 2)
        competition_type (str): Inherited from CompetitionBase.
        phase_type (str): Designation of the competition phase. Inherited from CompetitionPhaseBase.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    season_id: UUID | None = field(default=None)
    label: str | None = field(default=None, validator=validators.max_len(20))


# SQLAlchemy Imperative Mappings

phase = Table(
    "phase",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("season_id", SA_UUID, ForeignKey("season.id"), nullable=True),
    Column(
        "label",
        String(20),
        nullable=False,
        unique=True,
    ),
    Column("competition_type", String(30), nullable=True),
    Column("phase_type", String(30), nullable=True),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)


# ORM Relationships

mapper.map_imperatively(
    Phase,
    phase,
    properties={
        "season": relationship("Season", back_populates="phases"),
        "fixtures": relationship("Fixture", back_populates="phase"),
    },
)



from enum import Enum


class FixtureStatus(str, Enum):
    P = "Played"
    U = "Unplayed"
    F = "Forfeit"
    A = "Abandoned"
    D = "Postponed"


class FixtureResult(str, Enum):
    W = "Win"
    D = "Draw"
    L = "Loss"
    F = "Forfeit"
    S = "Suspended"
    N = "None"


class Category(str, Enum):
    """League categories."""

    MEN = "men"
    WOMEN = "women"
    COED = "coed"


class Division(str, Enum):
    """League divisions."""

    A = "A"
    B = "B"
    C = "C"
    D = "D"
    E = "E"
    F = "F"
    G = "G"
    H = "H"


class MatchDay(str, Enum):
    """Day of week."""

    MON = "Monday"
    TUE = "Tuesday"
    WED = "Wednesday"
    THU = "Thursday"
    FRI = "Friday"
    SAT = "Saturday"
    SUN = "Sunday"


class Gender(str, Enum):
    """Gender."""

    M = "Male"
    F = "Female"
    NB = "Non-Binary"
    T = "Transgender"
    O = "Other"  # noqa: E741
    P = "Prefer not to say"
    N = "N/A"


class Field(str, Enum):
    """Field of play."""

    A = "A"
    B = "B"
    C = "C"
    D = "D"
    E = "E"
    F = "F"
    G = "G"
    H = "H"
    I = "I"  # noqa: E741
    J = "J"
    K = "K"
    L = "L"


class ScheduleType(str, Enum):
    ROUNDROBIN = "Round Robin"
    TOURNAMENT = "Tournament"
    BRACKET = "Bracket"
    LADDER = "Ladder"



import pytest
from sqlalchemy import delete

from leaguemanager import models as m
from leaguemanager import services as s
from leaguemanager.core.toolbox import clear_table

org_data = {
    "name": "Test Organization",
}

league_data = {
    "name": "Test League",
}


season_data = {
    "name": "Season Test",
    "description": "Burndt",
    "projected_start_date": "2022-01-01",
}


team_data = [
    {"name": "Loro"},
    {"name": "June"},
    {"name": "Tripoli"},
    {"name": "Tres", "active": False},
]

athletes_data = [
    {"first_name": "Rob", "last_name": "Crow", "label": "Ace", "team": "Loro"},
    {"first_name": "Armistead", "last_name": "Smith IV", "label": "Bingo", "team": "Loro"},
    {"first_name": "Chris", "last_name": "Prescott", "label": "Rembrandt", "team": "Loro"},
]


@pytest.fixture(scope="module")
def org(org_service):
    return org_service.create(
        {
            "name": "Test Organization",
        },
        auto_commit=True,
    )


@pytest.fixture(scope="module")
def league():
    return m.League(**league_data)


@pytest.fixture(scope="module")
def properties(league_properties_service):
    return league_properties_service.create(
        {"sport": "Basketweaving", "category": "MEN", "division": "A"},
        auto_commit=True,
    )


@pytest.fixture(scope="module")
def ruleset(ruleset_service):
    return ruleset_service.create(
        {
            "number_of_games": 10,
            "sun_fixtures": True,
        },
        auto_commit=True,
    )


@pytest.fixture(scope="module")
def season():
    return m.Season(**season_data)


@pytest.fixture(scope="module")
def teams():
    return [m.Team(**data) for data in team_data]


def test_season_with_teams_active(
    session, season_service, league_service, team_member_service, team_service, org, league, season, teams
) -> None:
    new_league = league_service.create(league, auto_commit=True)
    new_league.organization_id = org.id
    new_season = season_service.create(season, auto_commit=True)
    new_season.league_id = new_league.id

    for team in teams:
        _team_memb = team_member_service.create(
            {"label": f"{team.name} member", "season_id": new_season.id}, auto_commit=True
        )
        team.team_membership_id = _team_memb.id
        _ = team_service.create(team, auto_commit=True)

    teams_in_season = season_service.all_teams(new_season.id)
    assert new_season.name == "Season Test"
    assert len(new_season.team_memberships) == len(teams)
    assert len(teams_in_season) == 3

    clear_table(session, m.League)
    clear_table(session, m.Season)
    clear_table(session, m.Team)
    clear_table(session, m.TeamMembership)


def test_season_with_teams_incl_inactive(
    session, season_service, league_service, team_member_service, team_service, org, league, season, teams
) -> None:
    new_league = league_service.create(league, auto_commit=True)
    new_league.organization_id = org.id
    new_season = season_service.create(season, auto_commit=True)
    new_season.league_id = new_league.id

    for team in teams:
        _team_memb = team_member_service.create(
            {"label": f"{team.name} member", "season_id": new_season.id}, auto_commit=True
        )
        team.team_membership_id = _team_memb.id
        _ = team_service.create(team, auto_commit=True)

    teams_in_season = season_service.all_teams(new_season.id, active=False)
    assert new_season.name == "Season Test"
    assert len(new_season.team_memberships) == len(teams)
    assert len(teams_in_season) == 4

    clear_table(session, m.League)
    clear_table(session, m.Season)
    clear_table(session, m.Team)
    clear_table(session, m.TeamMembership)


def test_season_all_athletes(
    session,
    season_service,
    league_service,
    team_member_service,
    team_service,
    athlete_service,
    individual_membership_service,
    org,
    league,
    season,
    teams,
) -> None:
    new_league = league_service.create(league, auto_commit=True)
    new_league.organization_id = org.id
    new_season = season_service.create(season, auto_commit=True)
    new_season.league_id = new_league.id

    athlete_service.create_many(athletes_data, auto_commit=True)

    rob = athlete_service.get_one_or_none(label="Ace")
    armistead = athlete_service.get_one_or_none(label="Bingo")
    chris = athlete_service.get_one_or_none(label="Rembrandt")

    for team in teams:
        _team_memb = team_member_service.create(
            {"label": f"{team.name} member", "season_id": new_season.id}, auto_commit=True
        )
        team.team_membership_id = _team_memb.id
        _ = team_service.create(team, auto_commit=True)

    loro_team = team_service.get_one_or_none(name="Loro")
    tres_team = team_service.get_one_or_none(name="Tres")

    individual_membership_service.create_many(
        [
            {"athlete_id": rob.id, "team_id": loro_team.id},
            {"athlete_id": armistead.id, "team_id": loro_team.id},
            {"athlete_id": chris.id, "team_id": tres_team.id},
        ],
        auto_commit=True,
    )

    athletes = season_service.all_athletes(new_season.id)

    assert new_season.name == "Season Test"
    assert len(new_season.team_memberships) == len(teams)
    assert len(teams) == 4
    assert len(athletes) == 2

    clear_table(session, m.League)
    clear_table(session, m.Season)
    clear_table(session, m.Team)
    clear_table(session, m.TeamMembership)
    clear_table(session, m.Athlete)
    clear_table(session, m.IndividualMembership)


def test_season_all_athletes_incl_inactive_team(
    session,
    season_service,
    league_service,
    team_member_service,
    team_service,
    athlete_service,
    individual_membership_service,
    org,
    league,
    season,
    teams,
) -> None:
    new_league = league_service.create(league, auto_commit=True)
    new_league.organization_id = org.id
    new_season = season_service.create(season, auto_commit=True)
    new_season.league_id = new_league.id

    athlete_service.create_many(athletes_data, auto_commit=True)

    rob = athlete_service.get_one_or_none(label="Ace")
    armistead = athlete_service.get_one_or_none(label="Bingo")
    chris = athlete_service.get_one_or_none(label="Rembrandt")

    for team in teams:
        _team_memb = team_member_service.create(
            {"label": f"{team.name} member", "season_id": new_season.id}, auto_commit=True
        )
        team.team_membership_id = _team_memb.id
        _ = team_service.create(team, auto_commit=True)

    loro_team = team_service.get_one_or_none(name="Loro")
    tres_team = team_service.get_one_or_none(name="Tres")

    individual_membership_service.create_many(
        [
            {"athlete_id": rob.id, "team_id": loro_team.id},
            {"athlete_id": armistead.id, "team_id": loro_team.id},
            {"athlete_id": chris.id, "team_id": tres_team.id},
        ],
        auto_commit=True,
    )

    athletes = season_service.all_athletes(new_season.id, incl_inactive_team=True)

    assert new_season.name == "Season Test"
    assert len(new_season.team_memberships) == len(teams)
    assert len(teams) == 4
    assert len(athletes) == 3

    clear_table(session, m.League)
    clear_table(session, m.Season)
    clear_table(session, m.Team)
    clear_table(session, m.TeamMembership)
    clear_table(session, m.Athlete)
    clear_table(session, m.IndividualMembership)



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import ManagerMembership
from leaguemanager.repository import ManagerMembershipSyncRepository
from leaguemanager.repository._async import ManagerMembershipAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["ManagerMembershipService", "ManagerMembershipAsyncService"]


class ManagerMembershipService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = ManagerMembershipSyncRepository


class ManagerMembershipAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = ManagerMembershipAsyncRepository



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Official
from leaguemanager.repository import OfficialSyncRepository
from leaguemanager.repository._async import OfficialAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["OfficialService", "OfficialAsyncService"]


class OfficialService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = OfficialSyncRepository


class OfficialAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = OfficialAsyncRepository



from __future__ import annotations

from datetime import UTC, datetime

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, String, Table
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import CompetitionBase, mapper, metadata

if TYPE_CHECKING:
    from uuid import UUID

    from .organization import Organization


@define(slots=False)
class League(CompetitionBase):
    """Defines the top level of a competition.

    This is ordinarily a child of the Organization (GoverningBody) and usually refers to
    a recurring competition type (i.e., Soccer League, Little League Baseball, etc...).
    Each League can have Season and/or other CompetitionBase-type objects as children.

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        organization_id (UUID): ForeignKey to Organization.
        competition_type (str): Describes competiton type. Inherited from CompetitionBase.
        name (str): Descriptive name. (i.e., Coed Hockey League, Flag Football at the Park, etc...)
        description (str): Optional description of the league.
        active (bool): Defaults to True. Can be used to deactivate this League.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    organization_id: UUID | None = field(default=None)
    name: str | None = field(default=None, validator=validators.max_len(80))
    description: str | None = field(default=None, validator=validators.optional(validators.max_len(120)))
    active: bool = field(default=True)


# SQLAlchemy Imperative Mappings

league = Table(
    "league",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("organization_id", SA_UUID, ForeignKey("organization.id"), nullable=True),
    Column("competition_type", String(30), default=("recurring-competition"), nullable=False),
    Column(
        "name",
        String(80),
        nullable=False,
        unique=True,
    ),
    Column("description", String(120), nullable=True),
    Column("active", Boolean, default=True),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships

mapper.map_imperatively(
    League,
    league,
    properties={
        "organization": relationship("Organization", back_populates="leagues"),
        "seasons": relationship("Season", back_populates="league"),
        "site": relationship("Site", back_populates="league"),
        "league_properties": relationship(
            "LeagueProperties", back_populates="league", uselist=False, cascade="all, delete-orphan"
        ),
    },
)



from .config import (
    async_alembic_config,
    async_config,
    async_engine,
    async_session_factory,
    get_async_session,
    get_session,
    sync_alembic_config,
    sync_config,
    sync_engine,
    sync_session_factory,
    validate_config,
)
from .sqlite3_datetime import register_sqlite

__all__ = [
    "async_alembic_config",
    "async_config",
    "async_engine",
    "async_session_factory",
    "get_async_session",
    "get_session",
    "register_sqlite",
    "sync_alembic_config",
    "sync_config",
    "sync_engine",
    "sync_session_factory",
    "validate_config",
]



from ._sync import (
    AddressSyncRepository,
    AthleteStatsSyncRepository,
    AthleteSyncRepository,
    CitySyncRepository,
    CountrySyncRepository,
    FixtureSyncRepository,
    IndividualMembershipSyncRepository,
    LeaguePropertiesSyncRepository,
    LeagueSyncRepository,
    ManagerMembershipSyncRepository,
    ManagerSyncRepository,
    ManagingSyncRepository,
    OfficialSyncRepository,
    OfficiatingSyncRepository,
    OrganizationSyncRepository,
    PhaseSyncRepository,
    RoleSyncRepository,
    RulesetSyncRepository,
    SeasonSyncRepository,
    SiteSyncRepository,
    StateSyncRepository,
    TeamMembershipSyncRepository,
    TeamStatsSyncRepository,
    TeamSyncRepository,
    UserRoleSyncRepository,
    UserSyncRepository,
)

__all__ = [
    "AddressSyncRepository",
    "AthleteSyncRepository",
    "AthleteStatsSyncRepository",
    "CitySyncRepository",
    "CountrySyncRepository",
    "FixtureSyncRepository",
    "IndividualMembershipSyncRepository",
    "LeagueSyncRepository",
    "LeaguePropertiesSyncRepository",
    "ManagerMembershipSyncRepository",
    "ManagerSyncRepository",
    "ManagingSyncRepository",
    "OfficialSyncRepository",
    "OfficiatingSyncRepository",
    "OrganizationSyncRepository",
    "PhaseSyncRepository",
    "RoleSyncRepository",
    "RulesetSyncRepository",
    "SeasonSyncRepository",
    "SiteSyncRepository",
    "StateSyncRepository",
    "TeamMembershipSyncRepository",
    "TeamStatsSyncRepository",
    "TeamSyncRepository",
    "UserRoleSyncRepository",
    "UserSyncRepository",
]



from datetime import datetime

import pytest

from leaguemanager.core.toolbox import clear_table
from leaguemanager.models import Fixture, League, Season, Team

season_data = {
    "name": "Season Scheduler Test",
    "description": "Burndt",
    "projected_start_date": "2022-01-01",
}

league_data = {
    "name": "League Scheduler Test",
}

schedule_data = {
    "name": "Schedule for Test Scheduler",
    "total_games": 10,
    "start_date": "2025-1-05 08:00:00",
}


@pytest.fixture(scope="module")
def season():
    return Season(**season_data)


@pytest.fixture(scope="module")
def league():
    return League(**league_data)


@pytest.fixture(scope="module")
def teams(all_teams):
    return [Team(**team_data) for team_data in all_teams]


# @pytest.fixture(scope="module")
# def sunday_schedule(season, league):
#     schedule = Schedule(
#         league_id=league.id,
#         season_id=season.id,
#         name="Schedule for Test Scheduler",
#         total_games=10,
#         start_date="2025-1-05 08:00:00",
#         concurrent_games=2,
#     )

#     return schedule


# @pytest.fixture(scope="module")
# def scheduler(session, sunday_schedule) -> Scheduler:
#     scheduler = Scheduler(session=session, schedule=sunday_schedule)
#     return scheduler


def test_generator_split_teams_even(scheduler, teams):
    split_teams = scheduler.generator.split_teams(teams)

    assert split_teams == (teams[:4], teams[4:])


def test_generator_split_teams_odd(scheduler, teams):
    temp_remove_team = teams.pop()

    assert len(teams) == 7

    with pytest.raises(ValueError):
        _ = scheduler.generator.split_teams(teams)

    teams.append(temp_remove_team)


@pytest.mark.parametrize("matchday", [1, 2, 3])
def test_generator_create_matchups_first_matchday(scheduler, matchday, teams):
    matchups = scheduler.generator.create_matchups(matchday=matchday, teams=teams)
    assert len(matchups) == 4


def test_increment_matchday(scheduler):
    matchday_1_start_time = scheduler.generator.increment_matchday(matchday=1)
    matchday_2_start_time = scheduler.generator.increment_matchday(matchday=2)

    assert matchday_1_start_time == datetime(2025, 1, 5, 8, 0)
    assert matchday_2_start_time == datetime(2025, 1, 12, 8, 0)


@pytest.mark.parametrize(
    "count, concurrent, expected",
    [(1, 2, "A"), (2, 2, "B"), (3, 4, "C"), (4, 4, "D"), (2, 1, "A"), (3, 1, "A")],
)
def test_determine_field(scheduler, count, concurrent, expected):
    field = scheduler.generator.determine_field(count, concurrent)
    assert field == expected


def test_home_or_away_even_round(scheduler, all_teams):
    fake_match = (all_teams[0], all_teams[1])

    home, away = scheduler.generator.home_or_away(round_number=2, match=fake_match)
    assert home == all_teams[0]
    assert away == all_teams[1]


def test_home_or_away_odd_round(scheduler, all_teams):
    fake_match = (all_teams[0], all_teams[1])

    home, away = scheduler.generator.home_or_away(round_number=1, match=fake_match)
    assert home == all_teams[1]
    assert away == all_teams[0]


@pytest.mark.parametrize(
    "_concurrent, _matchday, _match_count, expected",
    [
        (1, 1, 1, datetime(2025, 1, 5, 8, 0)),
        (1, 1, 2, datetime(2025, 1, 5, 10, 0)),
        (1, 1, 3, datetime(2025, 1, 5, 12, 0)),
        (1, 1, 4, datetime(2025, 1, 5, 14, 0)),
        (1, 1, 5, datetime(2025, 1, 5, 16, 0)),
        (1, 2, 1, datetime(2025, 1, 12, 8, 0)),
        (1, 2, 3, datetime(2025, 1, 12, 12, 0)),
        (1, 2, 5, datetime(2025, 1, 12, 16, 0)),
        (2, 1, 1, datetime(2025, 1, 5, 8, 0)),
        (2, 1, 2, datetime(2025, 1, 5, 8, 0)),
        (2, 1, 3, datetime(2025, 1, 5, 10, 0)),
        (2, 1, 4, datetime(2025, 1, 5, 10, 0)),
        (2, 1, 5, datetime(2025, 1, 5, 12, 0)),
        (2, 1, 6, datetime(2025, 1, 5, 12, 0)),
        (3, 2, 1, datetime(2025, 1, 12, 8, 0)),
        (3, 3, 2, datetime(2025, 1, 19, 8, 0)),
        (3, 4, 3, datetime(2025, 1, 26, 8, 0)),
        (3, 5, 4, datetime(2025, 2, 2, 10, 0)),
        (3, 6, 7, datetime(2025, 2, 9, 12, 0)),
        (4, 1, 1, datetime(2025, 1, 5, 8, 0)),
        (4, 1, 4, datetime(2025, 1, 5, 8, 0)),
    ],
)
def test_determine_start_time_third_matchday(
    scheduler, sunday_schedule, _concurrent, _matchday, _match_count, expected
):
    sunday_schedule.concurrent_games = _concurrent
    start_time = scheduler.generator.determine_start_time(matchday=_matchday, match_count=_match_count)
    assert start_time == expected


def test_create_matchday_fixtures(scheduler, teams):
    fixtures = scheduler.generator.create_matchday_fixtures(matchday=1, round_number=1, teams=teams)
    assert len(fixtures) == 4


@pytest.mark.parametrize(["total_games", "expected"], [(8, 32), (10, 32), (15, 32), (16, 64), (20, 64), (24, 96)])
def test_generate_fixtures_eight_teams(scheduler, total_games, expected, teams):
    scheduler.generator.schedule.total_games = total_games
    fixtures = scheduler.generator.generate_fixtures(teams=teams)
    assert len(fixtures) == expected

    fixtures = scheduler.generator.generate_fixtures(shuffle_order=False, teams=teams)
    assert len(fixtures) == expected
    fixtures.clear()


@pytest.mark.parametrize(["total_games", "expected"], [(4, 8), (8, 16), (10, 16)])
def test_generate_fixtures_four_teams(scheduler, total_games, expected, teams):
    teams = teams[:4]

    scheduler.generator.schedule.total_games = total_games
    fixtures = scheduler.generator.generate_fixtures(teams=teams)
    assert len(fixtures) == expected

    fixtures = scheduler.generator.generate_fixtures(shuffle_order=False, teams=teams)
    assert len(fixtures) == expected
    fixtures.clear()


def test_generate_season_fixtures_too_few_teams(scheduler, session, teams):
    scheduler.generator.schedule.total_games = 3
    with pytest.raises(ValueError):
        scheduler.generate_fixtures(teams=teams)
    scheduler.generator.schedule.total_games = 10


# def test_sort_fixtures_by_date(session, scheduler, season, league, teams):
#     _schedule = Schedule(**schedule_data, season_id=season.id, league_id=league.id)
#     scheduler = Scheduler(session=session, schedule=_schedule)

#     fixtures = scheduler.generate_fixtures(teams=teams)
#     _schedule.fixtures = fixtures

#     assert fixtures[0].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[-1].date == datetime(2025, 2, 23, 10, 0)

#     sorted_fixtures = scheduler.sort_fixtures(order_by_field="date", sort_order="desc")

#     assert sorted_fixtures[0].date == datetime(2025, 2, 23, 10, 0)
#     assert sorted_fixtures[-1].date == datetime(2025, 1, 5, 8, 0)


# def test_push_fixture_to_end_of_schedule(session, scheduler, league, season, teams):
#     clear_table(session, Fixture)

#     _schedule = Schedule(
#         league_id=league.id,
#         season_id=season.id,
#         name="Test Schedule",
#         total_games=10,
#         start_date="2025-1-05 08:00:00",
#     )
#     scheduler = Scheduler(session=session, schedule=_schedule)
#     _fixtures = scheduler.generator.generate_fixtures(teams=teams)
#     _schedule.fixtures = _fixtures

#     fixtures = scheduler.schedule.fixtures

#     assert len(fixtures) == 32
#     assert fixtures[0].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[0].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[-1].date == datetime(2025, 2, 23, 10, 0)

#     scheduler.push_fixture_to_end_of_schedule(fixtures[0])

#     assert len(fixtures) == 32
#     assert fixtures[0].date == datetime(2025, 3, 2, 8, 0)
#     assert fixtures[1].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[-1].date == datetime(2025, 2, 23, 10, 0)


# def test_push_matchday_to_end_of_schedule(session, scheduler, season, league, teams):
#     _schedule = Schedule(
#         league_id=league.id,
#         season_id=season.id,
#         name="Test Schedule",
#         total_games=10,
#         start_date="2025-1-05 08:00:00",
#     )

#     fixtures = scheduler.generator.generate_fixtures(teams=teams[:4])
#     _schedule.fixtures = fixtures
#     scheduler = Scheduler(session=session, schedule=_schedule)

#     scheduler.sort_fixtures()

#     assert fixtures[0].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[1].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[2].date == datetime(2025, 1, 12, 8, 0)
#     assert fixtures[3].date == datetime(2025, 1, 12, 8, 0)
#     assert fixtures[4].date == datetime(2025, 1, 19, 8, 0)
#     assert fixtures[5].date == datetime(2025, 1, 19, 8, 0)
#     assert fixtures[-1].date == datetime(2025, 2, 23, 8, 0)

#     scheduler.push_matchday_to_end_of_schedule(2)

#     assert fixtures[0].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[1].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[2].date == datetime(2025, 3, 2, 8, 0)
#     assert fixtures[3].date == datetime(2025, 3, 2, 8, 0)
#     assert fixtures[4].date == datetime(2025, 1, 19, 8, 0)
#     assert fixtures[5].date == datetime(2025, 1, 19, 8, 0)
#     assert fixtures[-1].date == datetime(2025, 2, 23, 8, 0)


# def test_push_all_fixtures_by_one_week(session, scheduler, league, season, teams):
#     _schedule = Schedule(
#         league_id=league.id,
#         season_id=season.id,
#         name="Test Schedule",
#         total_games=10,
#         start_date="2025-1-05 08:00:00",
#     )

#     fixtures = scheduler.generator.generate_fixtures(teams=teams[:4])
#     _schedule.fixtures = fixtures
#     scheduler = Scheduler(session, schedule=_schedule)

#     fixture = fixtures[4]  # Contains matchday 3 (1/19/2025)
#     scheduler.push_all_fixtures_by_one_week(fixture)
#     assert fixtures[0].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[1].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[2].date == datetime(2025, 1, 12, 8, 0)
#     assert fixtures[3].date == datetime(2025, 1, 12, 8, 0)
#     assert fixtures[4].date == datetime(2025, 1, 26, 8, 0)
#     assert fixtures[5].date == datetime(2025, 1, 26, 8, 0)
#     assert fixtures[6].date == datetime(2025, 2, 2, 8, 0)
#     assert fixtures[7].date == datetime(2025, 2, 2, 8, 0)
#     assert fixtures[-1].date == datetime(2025, 3, 2, 8, 0)



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Fixture
from leaguemanager.repository import FixtureSyncRepository
from leaguemanager.repository._async import FixtureAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["FixtureService", "FixtureAsyncService"]


class FixtureService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = FixtureSyncRepository


class FixtureAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = FixtureAsyncRepository



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import League
from leaguemanager.repository import LeagueSyncRepository
from leaguemanager.repository._async import LeagueAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["LeagueService", "LeagueAsyncService"]


class LeagueService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = LeagueSyncRepository


class LeagueAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = LeagueAsyncRepository



from __future__ import annotations

from typing import Any
from uuid import UUID

from advanced_alchemy.filters import FilterTypes
from sqlalchemy import select

from leaguemanager.models import Managing
from leaguemanager.repository import ManagingSyncRepository
from leaguemanager.repository._async import ManagingAsyncRepository
from leaguemanager.services.base import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

__all__ = ["ManagingService", "ManagingAsyncService"]


class ManagingService(SQLAlchemySyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = ManagingSyncRepository


class ManagingAsyncService(SQLAlchemyAsyncRepositoryService):
    """Handles database operations for roles data."""

    repository_type = ManagingAsyncRepository



from importlib import import_module
from importlib.util import module_from_spec, spec_from_file_location
from inspect import isclass
from pathlib import Path
from types import ModuleType
from typing import Any, Generator, Iterable

from advanced_alchemy.config import SQLAlchemyAsyncConfig, SQLAlchemySyncConfig
from advanced_alchemy.repository import SQLAlchemyAsyncRepository, SQLAlchemySyncRepository
from advanced_alchemy.service import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService
from attrs import define, field

from leaguemanager.core import get_settings

get_settings.cache_clear()
settings = get_settings()


@define
class DynamicObjectLoader:
    app_dir: Path = field()
    root_dir: Path = field()
    search_dir: Path = field()
    local_only: bool = field(default=False)

    @app_dir.default
    def _app_dir(self):
        return settings.LEAGUE_MANAGER_APP_DIR

    @root_dir.default
    def _root_dir(self):
        return settings.LEAGUE_MANAGER_ROOT_DIR

    @search_dir.default
    def _search_dir(self):
        return settings.LEAGUE_MANAGER_APP_DIR

    @classmethod
    def local_app(cls, app_dir: Path | None = None, root_dir: Path | None = None, search_dir: Path | None = None):
        """Creates a DynamicObjectLoader for a user's application

        Args:
            app_dir (Path | None, optional): The path to user's application. Defaults to None.
            root_dir (Path | None, optional): The path to the root of the user's application. Defaults to None.
            search_dir (Path | None, optional): The path to search for modules. Defaults to None.

        Returns:
            DynamicObjectLoader: A configured DynamicObjectLoader with defaults for user's application
        """
        if not app_dir:
            app_dir = settings.APP_DIR
        if not root_dir:
            root_dir = settings.APP_ROOT
        if not search_dir:
            search_dir = app_dir
        return cls(app_dir=app_dir, root_dir=root_dir, search_dir=search_dir, local_only=True)

    def _is_sync_repo(self, item: Any):
        """Checks if item is SQLAlchemySyncRepository class or subclass"""
        return isclass(item) and issubclass(item, SQLAlchemySyncRepository)

    def _is_async_repo(self, item: Any):
        """Checks if item is SQLAlchemyAsyncRepository class or subclass"""
        return isclass(item) and issubclass(item, SQLAlchemyAsyncRepository)

    def _is_sync_service(self, item: Any):
        """Checks if item is SQLAlchemySyncRepositoryService class or subclass."""
        return isclass(item) and issubclass(item, SQLAlchemySyncRepositoryService)

    def _is_async_service(self, item: Any):
        """Checks if item is SQLAlchemyAsyncRepositoryService class or subclass."""
        return isclass(item) and issubclass(item, SQLAlchemyAsyncRepositoryService)

    def _is_sync_config(self, item: Any):
        """Checks if item is a SQLAlchemySyncConfig.

        Using `isclass(item)` doesn't work here. May revisit at a later time.
        """
        return item.__class__ == SQLAlchemySyncConfig

    def _is_async_config(self, item: Any):
        """Checks if item is a SQLAlchemyAsyncConfig.

        Using `isclass(item)` doesn't work here. May revisit at a later time.
        """
        return item.__class__ == SQLAlchemyAsyncConfig

    def get_repositories(
        self, is_async: bool = False, **kwargs
    ) -> list[SQLAlchemySyncRepository | SQLAlchemyAsyncRepository]:
        """Returns all SQLAlchemySyncRepository or SQLAlchemyAsyncRepository classes dynamically.

        Args:
            is_async (bool, optional): If True, returns all SQLAlchemyAsyncRepository classes.
            Defaults to False.
        """
        if is_async:
            return self.load_objects(self.search_dir, self._is_async_repo)
        return self.load_objects(self.search_dir, self._is_sync_repo)

    def get_services(
        self, is_async: bool = False, **kwargs
    ) -> list[SQLAlchemySyncRepositoryService | SQLAlchemyAsyncRepositoryService]:
        """Returns all SQLAlchemySyncRepositoryService or SQLAlchemyAsyncRepositoryService classes dynamically.

        Args:
            is_async (bool, optional): If True, returns all SQLAlchemyAsyncRepositoryService classes.
            Defaults to False.
        """
        if is_async:
            return self.load_objects(self.search_dir, self._is_async_service)
        return self.load_objects(self.search_dir, self._is_sync_service)

    def get_configs(self, is_async: bool = False) -> list[SQLAlchemySyncConfig | SQLAlchemyAsyncConfig]:
        """Gets all SQLAlchemySyncConfig or SQLAlchemyAsyncConfig classes dynamically.

        If try_local_first is True, the search will start on the host application, and if a config
        item is found, it will load instead of league manager's default.

        Args:
            is_async (bool, optional): If True, returns all SQLAlchemyAsyncConfig classes.
            Defaults to False.
        """
        if is_async:
            return self.load_objects(self.search_dir, self._is_async_config, local_only=self.local_only)
        return self.load_objects(self.search_dir, self._is_sync_config, local_only=self.local_only)

    def load_objects(self, search_dir: Path, compare: callable, local_only: bool = False) -> Iterable[ModuleType]:
        """Looks through a list of objects and returns them based on a comparison function.

        If `local_only` is True, then it will only search the local directory for modules. Use this to
        override League Manager's default search location.

        Args:
            search_dir (Path): Directory name to search recursively
            compare (bool): Boolean comparison of all py files in `search_dir` directory
            local_only (bool): If True, only search the local directory for modules. Defaults to False.

        Returns:
            list: All modules that match the `compare` function.
        """
        items = []
        if local_only:
            _full_paths = self.full_paths(search_dir)
            items += self.collect_matching_objects(_full_paths, compare, is_dotted_path=False)
        else:
            _dotted_paths = self.dotted_paths(search_dir)
            items += self.collect_matching_objects(_dotted_paths, compare)
        return items

    def collect_matching_objects(
        self, modules: Iterable[str | Path], compare: callable, is_dotted_path: bool = True
    ) -> Iterable[ModuleType]:
        """Looks for all classes that match the `compare` function in a list of dotted paths.

        It expects a list of dotted paths to modules. It attempts to import each item, ignoring
        circular imports and modules that don't exist. If a module has an `__all__` attribute, it
        will only run the compare function on items in that attribute, and will only return unique
        items found.

        Args:
            modules (list[str]): List of dotted paths to modules
            compare (callable | bool): Boolean comparison of all py files in `search_dir` directory. If set
                to True, then all items in the module will be returned.
            is_dotted_path (bool, optional): Expects `modules` to be a list of dotted relative paths to modules (e.g. "app.app").
                If False, expects a list of full paths (e.g. /home/user/app/app.py). Defaults to True.

        Returns:
            list: All modules that match the `compare` function.
        """
        items = []
        for mod in modules:
            if "__init__" in str(mod):
                continue
            try:
                if not is_dotted_path:
                    _module = self.import_module_from_full_path(mod.stem, str(mod))
                else:
                    _module = import_module(mod)

            except AttributeError:
                # Ignore modules that create circular imports
                continue
            except ModuleNotFoundError:
                # Ignore modules that don't exist
                continue
            except Exception as e:
                print(f"Failed to import {mod} due to: {e}")
                return None
            if hasattr(_module, "__all__"):
                objs = [getattr(_module, obj) for obj in _module.__all__]
                if not objs:
                    print("No objects found")
                    continue
                items += [o for o in objs if compare(o) and o not in items]
        return items

    def dotted_paths(self, search_path: Path) -> Generator[str, None, None]:
        """Creates dotted paths to all .py files in `search_path` directory.

        Will exclude any files in directories in the `_exclude` list.

        Args:
            search_path (Path): Directory name to search recursively

        Returns:
            Generator[str]: Generator of dotted paths to all .py files in `search_path` directory
        """
        _exclude = [".venv", "venv", "tests", "test", "docs", "migrations", "__main__", "__init__"]
        for _file in search_path.rglob("**/*.py"):
            mod_path = str(_file.relative_to(self.root_dir)).replace("/", ".")
            if any(e in mod_path for e in _exclude):
                continue
            yield f"{mod_path[:-3]}"

    def full_paths(self, search_path: Path) -> Generator[Path, None, None]:
        """Creates full paths to all .py files in `search_path` directory.

        Will exclude any files in directories in the `_exclude` list.

        Args:
            search_path (Path): Directory name to search recursively

        Returns:
            Generator[Path]: Generator of full paths to all .py files in `search_path` directory
        """
        _exclude = ["tests", "test", "docs", "migrations", "__main__", "__init__"]
        for _file in search_path.rglob("**/*.py"):
            if any(e in str(_file) for e in _exclude):
                continue
            yield _file

    def import_module_from_full_path(self, module_name: str, full_path: str) -> Any:
        spec = spec_from_file_location(module_name, full_path)
        _module = module_from_spec(spec)
        spec.loader.exec_module(_module)
        return _module



from __future__ import annotations

from datetime import UTC, datetime

from attrs import define, field, validators
from sqlalchemy import UUID as SA_UUID
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy import Enum as SA_Enum
from sqlalchemy.orm import relationship
from typing_extensions import TYPE_CHECKING

from leaguemanager.models.base import PropertiesBase, mapper, metadata
from leaguemanager.models.enums import ScheduleType

if TYPE_CHECKING:
    from uuid import UUID

    from .organization import Organization


@define(slots=False)
class Ruleset(PropertiesBase):
    """Defines the ruleset specific to a Competition, such as a Season.

    As opposed to a competition's Properties, the Ruleset provides the specifics and format
    of each Event/Fixture. It defines the parameters of a specific Season that is useful
    for generating a schedule (i.e., start date, number of games per season, length of games).

    Note: Only "round robin" schedule_type is implemented so far.

    TODO: Implement "bracket" and "tournament" schedule_type schedulers

    Attributes:
        id (UUID): Inherited from UUIDAuditBase
        season_id (UUID): ForeignKey to League.
        site_id (UUID): ForeignKey to Site.
        schedule_type (str): Affects how games are scheduled and standings displayed (i.e, "round robin", "tournament")
        start_date (datetime): The date (and time) of first game.
        game_length (int): Total time of a Fixture (in minutes, including breaks).
        time_between (int): Time between end of last Fixture and start of next.
        number_of_teams (int): Active teams this Season.
        number_of_games (int): Number of total games per team (if "round robin")
        tiebreaker (str): Tie breaker rules.
        total_play_locations (int): In the case of multiple play areas (i.e., Field 4, Court B, etc..)
        mon_fixtures (bool): Whether Fixtures can be scheduled on Monday.
        tue_fixtures (bool): Whether Fixtures can be scheduled on Tuesday.
        wed_fixtures (bool): Whether Fixtures can be scheduled on Wednesday.
        thu_fixtures (bool): Whether Fixtures can be scheduled on Thursday.
        fri_fixtures (bool): Whether Fixtures can be scheduled on Friday.
        sat_fixtures (bool): Whether Fixtures can be scheduled on Saturday.
        sun_fixtures (bool): Whether Fixtures can be scheduled on Sunday.
        created_at (datetime): Inherited from UUIDAuditBase
        updated_at (datetime): Inherited from UUIDAuditBase
    """

    season_id: UUID | None = field(default=None)
    site_id: UUID | None = field(default=None)
    schedule_type: str | None = field(
        default=None, validator=validators.optional(validators.in_({e.name for e in ScheduleType}))
    )
    start_date: datetime | None = field(default=None)
    game_length: int | None = field(default=None)
    time_between: int | None = field(default=None)
    number_of_teams: int | None = field(default=None)
    number_of_games: int | None = field(default=None)
    tiebreaker: str | None = field(default=None, validator=validators.optional(validators.max_len(12)))
    total_play_locations: int | None = field(default=1)
    mon_fixtures: bool = field(default=False)
    tue_fixtures: bool = field(default=False)
    wed_fixtures: bool = field(default=False)
    thu_fixtures: bool = field(default=False)
    fri_fixtures: bool = field(default=False)
    sat_fixtures: bool = field(default=False)
    sun_fixtures: bool = field(default=False)


ruleset = Table(
    "ruleset",
    metadata,
    Column("id", SA_UUID, primary_key=True),
    Column("season_id", SA_UUID, ForeignKey("season.id"), nullable=True),
    Column("site_id", SA_UUID, ForeignKey("site.id"), nullable=True),
    Column("schedule_type", SA_Enum(ScheduleType), nullable=True),
    Column("start_date", DateTime, default=None, nullable=True),
    Column("game_length", Integer, nullable=True),
    Column("time_between", Integer, nullable=True),
    Column("number_of_teams", Integer, nullable=True),
    Column("number_of_games", Integer, nullable=True),
    Column("tiebreaker", String(12), nullable=True),
    Column("total_play_locations", Integer, nullable=False),
    Column("mon_fixtures", Boolean, default=False),
    Column("tue_fixtures", Boolean, default=False),
    Column("wed_fixtures", Boolean, default=False),
    Column("thu_fixtures", Boolean, default=False),
    Column("fri_fixtures", Boolean, default=False),
    Column("sat_fixtures", Boolean, default=False),
    Column("sun_fixtures", Boolean, default=False),
    Column("created_at", DateTime, default=lambda: datetime.now(UTC)),
    Column("updated_at", DateTime, default=lambda: datetime.now(UTC), onupdate=datetime.now(UTC)),
)

# ORM Relationships

mapper.map_imperatively(
    Ruleset,
    ruleset,
    properties={
        "season": relationship("Season", back_populates="ruleset", uselist=False),
        "site": relationship("Site", back_populates="ruleset"),
    },
)
